def graphite_LGM50_thermal_conductivity_ORegan2021(T):
    """
    Wet negative electrode thermal conductivity as a function of the temperature from
    [1].

    References
    ----------
    .. [1] Kieran O’Regan, Ferran Brosa Planella, W. Dhammika Widanage, and Emma
    Kendrick. "Thermal-electrochemical parametrisation of a lithium-ion battery:
    mapping Li concentration and temperature dependencies." Journal of the
    Electrochemical Society, submitted (2021).

    Parameters
    ----------
    T: :class:`pybamm.Symbol`
       Dimensional temperature

    Returns
    -------
    :class:`pybamm.Symbol`
       Thermal conductivity
    """

    lambda_wet = -2.61e-4 * T ** 2 + 0.1726 * T - 24.49

    return lambda_wet



from .base_external_circuit import BaseModel, LeadingOrderBaseModel
from .explicit_control_external_circuit import (
    ExplicitCurrentControl,
    ExplicitPowerControl,
    ExplicitResistanceControl,
    LeadingOrderExplicitCurrentControl,
)
from .function_control_external_circuit import (
    FunctionControl,
    VoltageFunctionControl,
    PowerFunctionControl,
    ResistanceFunctionControl,
    CCCVFunctionControl,
    LeadingOrderFunctionControl,
    LeadingOrderVoltageFunctionControl,
    LeadingOrderPowerFunctionControl,
)



#
# Base model for the external circuit
#
import pybamm


class BaseModel(pybamm.BaseSubModel):
    """Model to represent the behaviour of the external circuit."""

    def __init__(self, param, options):
        super().__init__(param, options=options)

    def get_fundamental_variables(self):
        Q_Ah = pybamm.standard_variables.Q_Ah
        variables = {"Discharge capacity [A.h]": Q_Ah}
        if self.options["calculate discharge energy"] == "true":
            Q_Wh = pybamm.standard_variables.Q_Wh
            variables.update({"Discharge energy [W.h]": Q_Wh})
        return variables

    def set_initial_conditions(self, variables):
        Q_Ah = variables["Discharge capacity [A.h]"]
        self.initial_conditions[Q_Ah] = pybamm.Scalar(0)
        if self.options["calculate discharge energy"] == "true":
            Q_Wh = variables["Discharge energy [W.h]"]
            self.initial_conditions[Q_Wh] = pybamm.Scalar(0)

    def set_rhs(self, variables):
        # ODE for discharge capacity
        Q_Ah = variables["Discharge capacity [A.h]"]
        I = variables["Current [A]"]

        self.rhs[Q_Ah] = I * self.param.timescale / 3600
        if self.options["calculate discharge energy"] == "true":
            Q_Wh = variables["Discharge energy [W.h]"]
            V = variables["Terminal voltage [V]"]
            self.rhs[Q_Wh] = I * V * self.param.timescale / 3600


class LeadingOrderBaseModel(BaseModel):
    """Model to represent the behaviour of the external circuit, at leading order."""

    def __init__(self, param, options):
        super().__init__(param, options)

    def get_fundamental_variables(self):
        Q_Ah = pybamm.Variable("Leading-order discharge capacity [A.h]")
        variables = {"Discharge capacity [A.h]": Q_Ah}
        if self.options["calculate discharge energy"] == "true":
            Q_Wh = pybamm.Variable("Leading-order discharge energy [W.h]")
            variables.update({"Discharge energy [W.h]": Q_Wh})
        return variables



#
# Class for single particle with polynomial concentration profile
#
import pybamm

from .base_fickian import BaseFickian


class XAveragedPolynomialProfile(BaseFickian):
    """
    Class for molar conservation in a single x-averaged particle employing Fick's law,
    with an assumed polynomial concentration profile in r. Model equations from [1]_.

    Parameters
    ----------
    param : parameter class
        The parameters to use for this submodel
    domain : str
        The domain of the model either 'Negative' or 'Positive'
    name : str
        The name of the polynomial approximation to be used. Can be "uniform
        profile", "quadratic profile" or "quartic profile".
    options: dict
        A dictionary of options to be passed to the model.
        See :class:`pybamm.BaseBatteryModel`

    References
    ----------
    .. [1] VR Subramanian, VD Diwakar and D Tapriyal. “Efficient Macro-Micro Scale
           Coupled Modeling of Batteries”. Journal of The Electrochemical Society,
           152(10):A2002-A2008, 2005

    **Extends:** :class:`pybamm.particle.BaseParticle`
    """

    def __init__(self, param, domain, name, options):
        super().__init__(param, domain, options)
        self.name = name

        pybamm.citations.register("Subramanian2005")

    def get_fundamental_variables(self):
        # For all orders we solve an equation for the average concentration
        if self.domain == "Negative":
            c_s_av = pybamm.standard_variables.c_s_n_av

        elif self.domain == "Positive":
            c_s_av = pybamm.standard_variables.c_s_p_av

        variables = {
            "Average " + self.domain.lower() + " particle concentration": c_s_av
        }

        # For the fourth order polynomial approximation we also solve an
        # equation for the average concentration gradient. Note: in the original
        # paper this quantity is referred to as the flux, but here we make the
        # distinction between the flux defined as N = -D*dc/dr and the concentration
        # gradient q = dc/dr
        if self.name == "quartic profile":
            if self.domain == "Negative":
                q_s_av = pybamm.standard_variables.q_s_n_av
            elif self.domain == "Positive":
                q_s_av = pybamm.standard_variables.q_s_p_av
            variables.update(
                {
                    "Average "
                    + self.domain.lower()
                    + " particle concentration gradient": q_s_av
                }
            )

        return variables

    def get_coupled_variables(self, variables):
        domain_param = self.domain_param

        c_s_av = variables["Average " + self.domain.lower() + " particle concentration"]
        T_av = variables["X-averaged " + self.domain.lower() + " electrode temperature"]

        D_eff_av = self._get_effective_diffusivity(c_s_av, T_av)
        i_boundary_cc = variables["Current collector current density"]
        a_av = variables[
            "X-averaged "
            + self.domain.lower()
            + " electrode surface area to volume ratio"
        ]
        sgn = 1 if self.domain == "Negative" else -1

        j_xav = sgn * i_boundary_cc / (a_av * self.domain_param.l)

        # Set surface concentration based on polynomial order
        if self.name == "uniform profile":
            # The concentration is uniform so the surface value is equal to
            # the average
            c_s_surf_xav = c_s_av
        elif self.name == "quadratic profile":
            # The surface concentration is computed from the average concentration
            # and boundary flux
            # Note 1: here we use the total average interfacial current for the single
            # particle. We explicitly write this as the current density divided by the
            # electrode thickness instead of getting the average current from the
            # interface submodel since the interface submodel requires the surface
            # concentration to be defined first to compute the exchange current density.
            # Explicitly writing out the average interfacial current here avoids
            # KeyErrors due to variables not being set in the "right" order.
            # Note 2: the concentration, c, inside the diffusion coefficient, D, here
            # should really be the surface value, but this requires solving a nonlinear
            # equation for c_surf (if the diffusion coefficient is nonlinear), adding
            # an extra algebraic equation to solve. For now, using the average c is an
            # ok approximation and means the SPM(e) still gives a system of ODEs rather
            # than DAEs.
            c_s_surf_xav = c_s_av - domain_param.C_diff * (
                j_xav / 5 / domain_param.a_R / domain_param.gamma / D_eff_av
            )
        elif self.name == "quartic profile":
            # The surface concentration is computed from the average concentration,
            # the average concentration gradient and the boundary flux (see notes
            # for the quadratic profile)
            q_s_av = variables[
                "Average " + self.domain.lower() + " particle concentration gradient"
            ]
            c_s_surf_xav = (
                c_s_av
                + 8 * q_s_av / 35
                - domain_param.C_diff
                * (j_xav / 35 / domain_param.a_R / domain_param.gamma / D_eff_av)
            )

        # Set concentration depending on polynomial order
        # Since c_s_xav doesn't depend on x, we need to define a spatial
        # variable r which only has "... particle" and "current
        # collector" as domains
        r = pybamm.SpatialVariable(
            "r_" + self.domain.lower()[0],
            domain=[f"{self.domain.lower()} particle"],
            auxiliary_domains={"secondary": "current collector"},
            coord_sys="spherical polar",
        )
        if self.name == "uniform profile":
            # The concentration is uniform
            A = pybamm.PrimaryBroadcast(c_s_av, [self.domain.lower() + " particle"])
            B = 0
            C = 0
        elif self.name == "quadratic profile":
            # The concentration is given by c = A + B*r**2
            A = pybamm.PrimaryBroadcast(
                (1 / 2) * (5 * c_s_av - 3 * c_s_surf_xav),
                [self.domain.lower() + " particle"],
            )
            B = pybamm.PrimaryBroadcast(
                (5 / 2) * (c_s_surf_xav - c_s_av), [self.domain.lower() + " particle"]
            )
            C = 0
        elif self.name == "quartic profile":
            # The concentration is given by c = A + B*r**2 + C*r**4
            A = pybamm.PrimaryBroadcast(
                39 * c_s_surf_xav / 4 - 3 * q_s_av - 35 * c_s_av / 4,
                [self.domain.lower() + " particle"],
            )
            B = pybamm.PrimaryBroadcast(
                -35 * c_s_surf_xav + 10 * q_s_av + 35 * c_s_av,
                [self.domain.lower() + " particle"],
            )
            C = pybamm.PrimaryBroadcast(
                105 * c_s_surf_xav / 4 - 7 * q_s_av - 105 * c_s_av / 4,
                [self.domain.lower() + " particle"],
            )

        c_s_xav = A + B * r ** 2 + C * r ** 4
        c_s = pybamm.SecondaryBroadcast(c_s_xav, [self.domain.lower() + " electrode"])
        c_s_surf = pybamm.PrimaryBroadcast(
            c_s_surf_xav, [self.domain.lower() + " electrode"]
        )

        # Set flux based on polynomial order
        T_xav = pybamm.PrimaryBroadcast(T_av, [self.domain.lower() + " particle"])
        D_eff_xav = self._get_effective_diffusivity(c_s_xav, T_xav)

        if self.name == "uniform profile":
            # The flux is zero since there is no concentration gradient
            N_s_xav = pybamm.FullBroadcastToEdges(
                0, self.domain.lower() + " particle", "current collector"
            )
        elif self.name == "quadratic profile":
            # The flux may be computed directly from the polynomial for c
            N_s_xav = -D_eff_xav * 5 * (c_s_surf_xav - c_s_av) * r
        elif self.name == "quartic profile":
            q_s_av = variables[
                "Average " + self.domain.lower() + " particle concentration gradient"
            ]
            # The flux may be computed directly from the polynomial for c
            N_s_xav = -D_eff_xav * (
                (-70 * c_s_surf_xav + 20 * q_s_av + 70 * c_s_av) * r
                + (105 * c_s_surf_xav - 28 * q_s_av - 105 * c_s_av) * r ** 3
            )

        D_eff = pybamm.SecondaryBroadcast(
            D_eff_xav, [self._domain.lower() + " electrode"]
        )
        N_s = pybamm.SecondaryBroadcast(N_s_xav, [self._domain.lower() + " electrode"])

        variables.update(
            self._get_standard_concentration_variables(
                c_s, c_s_av=c_s_av, c_s_surf=c_s_surf
            )
        )
        variables.update(self._get_standard_flux_variables(N_s, N_s_xav))
        variables.update(self._get_standard_diffusivity_variables(D_eff))
        variables.update(self._get_total_concentration_variables(variables))

        return variables

    def set_rhs(self, variables):
        # Note: we have to use `pybamm.source(rhs, var)` in the rhs dict so that
        # the scalar source term gets multplied by the correct mass matrix when
        # using this model with 2D current collectors with the finite element
        # method (see #1399)
        domain_param = self.domain_param

        c_s_av = variables["Average " + self.domain.lower() + " particle concentration"]
        j_xav = variables[
            "X-averaged "
            + self.domain.lower()
            + " electrode interfacial current density"
        ]

        self.rhs = {
            c_s_av: pybamm.source(
                -3 * j_xav / domain_param.a_R / domain_param.gamma, c_s_av
            )
        }

        if self.name == "quartic profile":
            # We solve an extra ODE for the average particle concentration gradient
            q_s_av = variables[
                "Average " + self.domain.lower() + " particle concentration gradient"
            ]
            D_eff_xav = variables[
                "X-averaged " + self.domain.lower() + " effective diffusivity"
            ]

            self.rhs.update(
                {
                    q_s_av: pybamm.source(
                        -30 * pybamm.surf(D_eff_xav) * q_s_av / domain_param.C_diff
                        - 45 * j_xav / domain_param.a_R / domain_param.gamma / 2,
                        q_s_av,
                    )
                }
            )

    def set_initial_conditions(self, variables):
        """
        For single or x-averaged particle models, initial conditions can't depend on x
        or r so we take the r- and x-average of the initial conditions.
        """
        c_s_av = variables["Average " + self.domain.lower() + " particle concentration"]

        c_init = pybamm.x_average(pybamm.r_average(self.domain_param.c_init))

        self.initial_conditions = {c_s_av: c_init}
        if self.name == "quartic profile":
            # We also need to provide an initial condition for the average
            # concentration gradient
            q_s_av = variables[
                "Average " + self.domain.lower() + " particle concentration gradient"
            ]
            self.initial_conditions.update({q_s_av: 0})



#
# Class for varying active material volume fraction, driven by stress
#
import pybamm

from .base_active_material import BaseModel


class LossActiveMaterial(BaseModel):
    """Submodel for varying active material volume fraction from [1]_ and [2]_.

    Parameters
    ----------
    param : parameter class
        The parameters to use for this submodel
    domain : str
        The domain of the model either 'Negative' or 'Positive'
    options : dict
        Additional options to pass to the model
    x_average : bool
        Whether to use x-averaged variables (SPM, SPMe, etc) or full variables (DFN)

    **Extends:** :class:`pybamm.active_material.BaseModel`

    References
    ----------
    .. [1] Ai, W., Kraft, L., Sturm, J., Jossen, A., & Wu, B. (2019). Electrochemical
           Thermal-Mechanical Modelling of Stress Inhomogeneity in Lithium-Ion Pouch
           Cells. Journal of The Electrochemical Society, 167(1), 013512.
    .. [2] Reniers, J. M., Mulder, G., & Howey, D. A. (2019). Review and performance
           comparison of mechanical-chemical degradation models for lithium-ion
           batteries. Journal of The Electrochemical Society, 166(14), A3189.
    """

    def __init__(self, param, domain, options, x_average):
        super().__init__(param, domain, options=options)
        pybamm.citations.register("Reniers2019")
        self.x_average = x_average

    def get_fundamental_variables(self):
        domain = self.domain.lower() + " electrode"
        if self.x_average is True:
            eps_solid_xav = pybamm.Variable(
                "X-averaged " + domain + " active material volume fraction",
                domain="current collector",
            )
            eps_solid = pybamm.PrimaryBroadcast(eps_solid_xav, domain)
        else:
            eps_solid = pybamm.Variable(
                self.domain + " electrode active material volume fraction",
                domain=domain,
                auxiliary_domains={"secondary": "current collector"},
            )
        variables = self._get_standard_active_material_variables(eps_solid)
        return variables

    def get_coupled_variables(self, variables):
        deps_solid_dt = 0
        lam_option = getattr(self.options, self.domain.lower())[
            "loss of active material"
        ]
        if "stress" in lam_option:
            # obtain the rate of loss of active materials (LAM) by stress
            # This is loss of active material model by mechanical effects
            if self.x_average is True:
                stress_t_surf = variables[
                    "X-averaged "
                    + self.domain.lower()
                    + " particle surface tangential stress"
                ]
                stress_r_surf = variables[
                    "X-averaged "
                    + self.domain.lower()
                    + " particle surface radial stress"
                ]
            else:
                stress_t_surf = variables[
                    self.domain + " particle surface tangential stress"
                ]
                stress_r_surf = variables[
                    self.domain + " particle surface radial stress"
                ]

            beta_LAM = self.domain_param.beta_LAM
            stress_critical = self.domain_param.stress_critical
            m_LAM = self.domain_param.m_LAM

            stress_h_surf = (stress_r_surf + 2 * stress_t_surf) / 3
            # compressive stress make no contribution
            stress_h_surf *= stress_h_surf > 0
            # assuming the minimum hydrostatic stress is zero for full cycles
            stress_h_surf_min = stress_h_surf * 0
            j_stress_LAM = (
                -beta_LAM
                * ((stress_h_surf - stress_h_surf_min) / stress_critical) ** m_LAM
            )
            deps_solid_dt += j_stress_LAM

        if "reaction" in lam_option:
            if self.x_average is True:
                a = variables[
                    "X-averaged "
                    + self.domain.lower()
                    + " electrode surface area to volume ratio"
                ]
            else:
                a = variables[self.domain + " electrode surface area to volume ratio"]

            beta_LAM_sei = self.domain_param.beta_LAM_sei
            if self.domain == "Negative":
                if self.x_average is True:
                    j_sei = variables["X-averaged SEI interfacial current density"]
                else:
                    j_sei = variables["SEI interfacial current density"]
            else:
                # No SEI in the positive electrode so no reaction-driven LAM
                # until other reactions are implemented
                j_sei = 0

            j_stress_reaction = beta_LAM_sei * a * j_sei
            deps_solid_dt += j_stress_reaction
        variables.update(
            self._get_standard_active_material_change_variables(deps_solid_dt)
        )
        return variables

    def set_rhs(self, variables):
        Domain = self.domain + " electrode"
        if self.x_average is True:
            eps_solid = variables[
                "X-averaged " + Domain.lower() + " active material volume fraction"
            ]
            deps_solid_dt = variables[
                "X-averaged "
                + Domain.lower()
                + " active material volume fraction change"
            ]
        else:
            eps_solid = variables[Domain + " active material volume fraction"]
            deps_solid_dt = variables[
                Domain + " active material volume fraction change"
            ]

        self.rhs = {eps_solid: deps_solid_dt}

    def set_initial_conditions(self, variables):

        eps_solid_init = self.domain_param.epsilon_s

        if self.x_average is True:
            eps_solid_xav = variables[
                "X-averaged "
                + self.domain.lower()
                + " electrode active material volume fraction"
            ]
            self.initial_conditions = {eps_solid_xav: pybamm.x_average(eps_solid_init)}
        else:
            eps_solid = variables[
                self.domain + " electrode active material volume fraction"
            ]
            self.initial_conditions = {eps_solid: eps_solid_init}



#
# Standard parameters for lithium-ion battery models
#
import pybamm
from .base_parameters import BaseParameters


class LithiumIonParameters(BaseParameters):
    """
    Standard parameters for lithium-ion battery models

    Layout:
        1. Dimensional Parameters
        2. Dimensional Functions
        3. Scalings
        4. Dimensionless Parameters
        5. Dimensionless Functions
        6. Input Current

    Parameters
    ----------

    options : dict, optional
        A dictionary of options to be passed to the parameters. The options that
        can be set are listed below.

            * "particle shape" : str, optional
                Sets the model shape of the electrode particles. This is used to
                calculate the surface area to volume ratio. Can be "spherical"
                (default). TODO: implement "cylindrical" and "platelet".
            * "working electrode": str
                Which electrode(s) intercalates and which is counter. If "both"
                (default), the model is a standard battery. Otherwise can be "negative"
                or "positive" to indicate a half-cell model.

    """

    def __init__(self, options=None):
        self.options = options
        # Save whether the submodel is a half-cell submodel
        self.half_cell = self.options["working electrode"] != "both"

        # Get geometric, electrical and thermal parameters
        self.geo = pybamm.geometric_parameters
        self.elec = pybamm.electrical_parameters
        self.therm = pybamm.thermal_parameters

        # Spatial variables
        r_n = pybamm.standard_spatial_vars.r_n * self.geo.n.R_typ
        r_p = pybamm.standard_spatial_vars.r_p * self.geo.p.R_typ
        x_n = pybamm.standard_spatial_vars.x_n * self.geo.L_x
        x_s = pybamm.standard_spatial_vars.x_s * self.geo.L_x
        x_p = pybamm.standard_spatial_vars.x_p * self.geo.L_x

        # Initialize domain parameters
        self.n = DomainLithiumIonParameters("Negative", self, x_n, r_n)
        self.s = DomainLithiumIonParameters("Separator", self, x_s, None)
        self.p = DomainLithiumIonParameters("Positive", self, x_p, r_p)
        self.domain_params = [self.n, self.s, self.p]

        # Set parameters and scales
        self._set_dimensional_parameters()
        self._set_scales()
        self._set_dimensionless_parameters()

        # Set input current
        self._set_input_current()

    def _set_dimensional_parameters(self):
        """Defines the dimensional parameters"""
        # Physical constants
        self.R = pybamm.constants.R
        self.F = pybamm.constants.F
        self.k_b = pybamm.constants.k_b
        self.q_e = pybamm.constants.q_e

        self.T_ref = self.therm.T_ref
        self.T_init_dim = self.therm.T_init_dim
        self.T_init = self.therm.T_init

        # Macroscale geometry
        self.L_x = self.geo.L_x
        self.L = self.geo.L
        self.L_y = self.geo.L_y
        self.L_z = self.geo.L_z
        self.r_inner_dimensional = self.geo.r_inner_dimensional
        self.r_outer_dimensional = self.geo.r_outer_dimensional
        self.A_cc = self.geo.A_cc
        self.A_cooling = self.geo.A_cooling
        self.V_cell = self.geo.V_cell

        # Electrical
        self.I_typ = self.elec.I_typ
        self.Q = self.elec.Q
        self.C_rate = self.elec.C_rate
        self.n_electrodes_parallel = self.elec.n_electrodes_parallel
        self.n_cells = self.elec.n_cells
        self.i_typ = self.elec.i_typ
        self.voltage_low_cut_dimensional = self.elec.voltage_low_cut_dimensional
        self.voltage_high_cut_dimensional = self.elec.voltage_high_cut_dimensional

        # Domain parameters
        for domain in self.domain_params:
            domain._set_dimensional_parameters()

        # Electrolyte properties
        self.c_e_typ = pybamm.Parameter("Typical electrolyte concentration [mol.m-3]")

        if self.half_cell:
            self.epsilon_init = pybamm.concatenation(
                self.s.epsilon_init, self.p.epsilon_init
            )
        else:
            self.epsilon_init = pybamm.concatenation(
                self.n.epsilon_init, self.s.epsilon_init, self.p.epsilon_init
            )

        # SEI parameters
        self.V_bar_inner_dimensional = pybamm.Parameter(
            "Inner SEI partial molar volume [m3.mol-1]"
        )
        self.V_bar_outer_dimensional = pybamm.Parameter(
            "Outer SEI partial molar volume [m3.mol-1]"
        )

        self.m_sei_dimensional = pybamm.Parameter(
            "SEI reaction exchange current density [A.m-2]"
        )

        self.R_sei_dimensional = pybamm.Parameter("SEI resistivity [Ohm.m]")
        self.D_sol_dimensional = pybamm.Parameter(
            "Outer SEI solvent diffusivity [m2.s-1]"
        )
        self.c_sol_dimensional = pybamm.Parameter(
            "Bulk solvent concentration [mol.m-3]"
        )
        self.m_ratio = pybamm.Parameter(
            "Ratio of inner and outer SEI exchange current densities"
        )
        self.U_inner_dimensional = pybamm.Parameter(
            "Inner SEI open-circuit potential [V]"
        )
        self.U_outer_dimensional = pybamm.Parameter(
            "Outer SEI open-circuit potential [V]"
        )
        self.kappa_inner_dimensional = pybamm.Parameter(
            "Inner SEI electron conductivity [S.m-1]"
        )
        self.D_li_dimensional = pybamm.Parameter(
            "Inner SEI lithium interstitial diffusivity [m2.s-1]"
        )
        self.c_li_0_dimensional = pybamm.Parameter(
            "Lithium interstitial reference concentration [mol.m-3]"
        )
        self.L_inner_0_dim = pybamm.Parameter("Initial inner SEI thickness [m]")
        self.L_outer_0_dim = pybamm.Parameter("Initial outer SEI thickness [m]")
        self.L_sei_0_dim = self.L_inner_0_dim + self.L_outer_0_dim

        # EC reaction
        self.c_ec_0_dim = pybamm.Parameter(
            "EC initial concentration in electrolyte [mol.m-3]"
        )
        self.D_ec_dim = pybamm.Parameter("EC diffusivity [m2.s-1]")
        self.k_sei_dim = pybamm.Parameter("SEI kinetic rate constant [m.s-1]")
        self.U_sei_dim = pybamm.Parameter("SEI open-circuit potential [V]")

        # Li plating parameters
        self.V_bar_plated_Li = pybamm.Parameter(
            "Lithium metal partial molar volume [m3.mol-1]"
        )
        self.c_plated_Li_0_dim = pybamm.Parameter(
            "Initial plated lithium concentration [mol.m-3]"
        )

        # Initial conditions
        # Note: the initial concentration in the electrodes can be set as a function
        # of through-cell position, so is defined later as a function
        self.c_e_init_dimensional = pybamm.Parameter(
            "Initial concentration in electrolyte [mol.m-3]"
        )

        self.alpha_T_cell_dim = pybamm.Parameter(
            "Cell thermal expansion coefficient [m.K-1]"
        )

        # Total lithium
        c_e_av_init = pybamm.xyz_average(self.epsilon_init) * self.c_e_typ
        self.n_Li_e_init = c_e_av_init * self.L_x * self.A_cc

        self.n_Li_particles_init = self.n.n_Li_init + self.p.n_Li_init
        self.n_Li_init = self.n_Li_particles_init + self.n_Li_e_init

        # Reference OCP based on initial concentration
        self.ocv_ref = self.p.U_ref - self.n.U_ref
        self.ocv_init_dim = self.p.U_init_dim - self.n.U_init_dim

    def D_e_dimensional(self, c_e, T):
        """Dimensional diffusivity in electrolyte"""
        inputs = {"Electrolyte concentration [mol.m-3]": c_e, "Temperature [K]": T}
        return pybamm.FunctionParameter("Electrolyte diffusivity [m2.s-1]", inputs)

    def kappa_e_dimensional(self, c_e, T):
        """Dimensional electrolyte conductivity"""
        inputs = {"Electrolyte concentration [mol.m-3]": c_e, "Temperature [K]": T}
        return pybamm.FunctionParameter("Electrolyte conductivity [S.m-1]", inputs)

    def j0_stripping_dimensional(self, c_e, c_Li, T):
        """Dimensional exchange-current density for stripping [A.m-2]"""
        inputs = {
            "Electrolyte concentration [mol.m-3]": c_e,
            "Plated lithium concentration [mol.m-3]": c_Li,
            "Temperature [K]": T,
        }
        return pybamm.FunctionParameter(
            "Exchange-current density for stripping [A.m-2]", inputs
        )

    def j0_plating_dimensional(self, c_e, c_Li, T):
        """Dimensional exchange-current density for plating [A.m-2]"""
        inputs = {
            "Electrolyte concentration [mol.m-3]": c_e,
            "Plated lithium concentration [mol.m-3]": c_Li,
            "Temperature [K]": T,
        }
        return pybamm.FunctionParameter(
            "Exchange-current density for plating [A.m-2]", inputs
        )

    def _set_scales(self):
        """Define the scales used in the non-dimensionalisation scheme"""
        # Concentration
        self.electrolyte_concentration_scale = self.c_e_typ

        # Electrical
        # Both potential scales are the same but they have different units
        self.potential_scale = self.R * self.T_ref / self.F  # volts
        self.potential_scale_eV = self.k_b / self.q_e * self.T_ref  # eV
        self.current_scale = self.i_typ
        self.current_scale.print_name = "I_typ"

        # Thermal
        self.Delta_T = self.therm.Delta_T

        # Velocity scale
        self.velocity_scale = pybamm.Scalar(1)

        # Discharge timescale
        if self.options["working electrode"] == "positive":
            self.c_max = self.p.c_max
        else:
            self.c_max = self.n.c_max
        self.tau_discharge = self.F * self.c_max * self.L_x / self.i_typ

        # Electrolyte diffusion timescale
        self.D_e_typ = self.D_e_dimensional(self.c_e_typ, self.T_ref)
        self.tau_diffusion_e = self.L_x ** 2 / self.D_e_typ

        # Thermal diffusion timescale
        self.tau_th_yz = self.therm.tau_th_yz

        # Choose discharge timescale
        if self.options["timescale"] == "default":
            self.timescale = self.tau_discharge
        else:
            self.timescale = pybamm.Scalar(self.options["timescale"])

        for domain in self.domain_params:
            domain._set_scales()

    def _set_dimensionless_parameters(self):
        """Defines the dimensionless parameters"""
        # Timescale ratios
        self.C_e = self.tau_diffusion_e / self.timescale
        self.C_th = self.tau_th_yz / self.timescale

        # Concentration ratios
        self.gamma_e = (self.tau_discharge / self.timescale) * self.c_e_typ / self.c_max

        # Macroscale Geometry
        self.l_x = self.geo.l_x
        self.l_y = self.geo.l_y
        self.l_z = self.geo.l_z
        self.r_inner = self.geo.r_inner
        self.r_outer = self.geo.r_outer
        self.a_cc = self.geo.a_cc
        self.a_cooling = self.geo.a_cooling
        self.v_cell = self.geo.v_cell
        self.l = self.geo.l
        self.delta = self.geo.delta

        for domain in self.domain_params:
            domain._set_dimensionless_parameters()

        # Electrolyte Properties
        self.beta_surf = pybamm.Scalar(0)

        # Electrical
        self.voltage_low_cut = (
            self.voltage_low_cut_dimensional - self.ocv_ref
        ) / self.potential_scale
        self.voltage_high_cut = (
            self.voltage_high_cut_dimensional - self.ocv_ref
        ) / self.potential_scale

        # Thermal
        self.Theta = self.therm.Theta
        self.T_amb_dim = self.therm.T_amb_dim
        self.T_amb = self.therm.T_amb

        self.h_edge = self.therm.h_edge
        self.h_total = self.therm.h_total
        self.rho = self.therm.rho

        self.B = (
            self.i_typ
            * self.R
            * self.T_ref
            * self.tau_th_yz
            / (self.therm.rho_eff_dim_ref * self.F * self.Delta_T * self.L_x)
        )

        # SEI parameters
        self.C_sei_reaction = (self.n.j_scale / self.m_sei_dimensional) * pybamm.exp(
            -(self.F * self.n.U_ref / (2 * self.R * self.T_ref))
        )

        self.C_sei_solvent = (
            self.n.j_scale
            * self.L_sei_0_dim
            / (self.c_sol_dimensional * self.F * self.D_sol_dimensional)
        )

        self.C_sei_electron = (
            self.n.j_scale
            * self.F
            * self.L_sei_0_dim
            / (self.kappa_inner_dimensional * self.R * self.T_ref)
        )

        self.C_sei_inter = (
            self.n.j_scale
            * self.L_sei_0_dim
            / (self.D_li_dimensional * self.c_li_0_dimensional * self.F)
        )

        self.U_inner_electron = self.F * self.U_inner_dimensional / self.R / self.T_ref

        self.R_sei = (
            self.F
            * self.n.j_scale
            * self.R_sei_dimensional
            * self.L_sei_0_dim
            / self.R
            / self.T_ref
        )

        self.v_bar = self.V_bar_outer_dimensional / self.V_bar_inner_dimensional
        self.c_sei_scale = (
            self.L_sei_0_dim * self.n.a_typ / self.V_bar_inner_dimensional
        )
        self.c_sei_outer_scale = (
            self.L_sei_0_dim * self.n.a_typ / self.V_bar_outer_dimensional
        )

        self.L_inner_0 = self.L_inner_0_dim / self.L_sei_0_dim
        self.L_outer_0 = self.L_outer_0_dim / self.L_sei_0_dim

        # ratio of SEI reaction scale to intercalation reaction
        self.Gamma_SEI = (
            self.V_bar_inner_dimensional * self.n.j_scale * self.timescale
        ) / (self.F * self.L_sei_0_dim)

        # EC reaction
        self.C_ec = (
            self.L_sei_0_dim
            * self.n.j_scale
            / (self.F * self.c_ec_0_dim * self.D_ec_dim)
        )
        self.C_sei_ec = (
            self.F
            * self.k_sei_dim
            * self.c_ec_0_dim
            / self.n.j_scale
            * (
                pybamm.exp(
                    -(
                        self.F
                        * (self.n.U_ref - self.U_sei_dim)
                        / (2 * self.R * self.T_ref)
                    )
                )
            )
        )
        self.beta_sei = self.n.a_typ * self.L_sei_0_dim * self.Gamma_SEI
        self.c_sei_init = self.c_ec_0_dim / self.c_sei_outer_scale

        # lithium plating parameters
        self.c_Li_typ = pybamm.Parameter(
            "Typical plated lithium concentration [mol.m-3]"
        )
        self.c_plated_Li_0 = self.c_plated_Li_0_dim / self.c_Li_typ

        # ratio of lithium plating reaction scaled to intercalation reaction
        self.Gamma_plating = (self.n.a_typ * self.n.j_scale * self.timescale) / (
            self.F * self.c_Li_typ
        )

        self.beta_plating = self.Gamma_plating * self.V_bar_plated_Li * self.c_Li_typ

        # Initial conditions
        self.c_e_init = self.c_e_init_dimensional / self.c_e_typ
        self.ocv_init = (self.ocv_init_dim - self.ocv_ref) / self.potential_scale

        # Dimensionless mechanical parameters
        self.t0_cr = 3600 / self.C_rate / self.timescale

    def chi(self, c_e, T):
        """
        Thermodynamic factor:
            (1-2*t_plus) is for Nernst-Planck,
            2*(1-t_plus) for Stefan-Maxwell,
        see Bizeray et al (2016) "Resolving a discrepancy ...".
        """
        return (2 * (1 - self.t_plus(c_e, T))) * (self.one_plus_dlnf_dlnc(c_e, T))

    def t_plus(self, c_e, T):
        """Cation transference number (dimensionless)"""
        inputs = {
            "Electrolyte concentration [mol.m-3]": c_e * self.c_e_typ,
            "Temperature [K]": self.Delta_T * T + self.T_ref,
        }
        return pybamm.FunctionParameter("Cation transference number", inputs)

    def one_plus_dlnf_dlnc(self, c_e, T):
        """Thermodynamic factor (dimensionless)"""
        inputs = {
            "Electrolyte concentration [mol.m-3]": c_e * self.c_e_typ,
            "Temperature [K]": self.Delta_T * T + self.T_ref,
        }
        return pybamm.FunctionParameter("1 + dlnf/dlnc", inputs)

    def D_e(self, c_e, T):
        """Dimensionless electrolyte diffusivity"""
        c_e_dimensional = c_e * self.c_e_typ
        T_dim = self.Delta_T * T + self.T_ref
        return self.D_e_dimensional(c_e_dimensional, T_dim) / self.D_e_typ

    def kappa_e(self, c_e, T):
        """Dimensionless electrolyte conductivity"""
        c_e_dimensional = c_e * self.c_e_typ
        kappa_scale = self.F ** 2 * self.D_e_typ * self.c_e_typ / (self.R * self.T_ref)
        T_dim = self.Delta_T * T + self.T_ref
        return self.kappa_e_dimensional(c_e_dimensional, T_dim) / kappa_scale

    def j0_stripping(self, c_e, c_Li, T):
        """Dimensionless exchange-current density for stripping"""
        c_e_dim = c_e * self.c_e_typ
        c_Li_dim = c_Li * self.c_Li_typ
        T_dim = self.Delta_T * T + self.T_ref

        return self.j0_stripping_dimensional(c_e_dim, c_Li_dim, T_dim) / self.n.j_scale

    def j0_plating(self, c_e, c_Li, T):
        """Dimensionless reverse plating current"""
        c_e_dim = c_e * self.c_e_typ
        c_Li_dim = c_Li * self.c_Li_typ
        T_dim = self.Delta_T * T + self.T_ref

        return self.j0_plating_dimensional(c_e_dim, c_Li_dim, T_dim) / self.n.j_scale

    def _set_input_current(self):
        """Set the input current"""

        self.dimensional_current_with_time = pybamm.FunctionParameter(
            "Current function [A]", {"Time [s]": pybamm.t * self.timescale}
        )
        self.dimensional_current_density_with_time = (
            self.dimensional_current_with_time
            / (self.n_electrodes_parallel * self.geo.A_cc)
        )
        self.current_with_time = (
            self.dimensional_current_with_time / self.I_typ * pybamm.sign(self.I_typ)
        )

    @property
    def options(self):
        return self._options

    @options.setter
    def options(self, extra_options):
        self._options = pybamm.BatteryModelOptions(extra_options)


class DomainLithiumIonParameters(BaseParameters):
    def __init__(self, domain, main_param, x, r):
        self.domain = domain
        self.main_param = main_param

        self.geo = getattr(main_param.geo, domain.lower()[0])
        self.therm = getattr(main_param.therm, domain.lower()[0])

        self.x = x
        self.r = r

    def _set_dimensional_parameters(self):
        main = self.main_param
        Domain = self.domain
        domain = Domain.lower()
        x = self.x
        r = self.r

        if Domain == "Separator":
            self.epsilon_init = pybamm.FunctionParameter(
                "Separator porosity", {"Through-cell distance (x) [m]": x}
            )
            self.epsilon_inactive = 1 - self.epsilon_init
            self.b_e = self.geo.b_e
            self.L = self.geo.L
            return

        # Macroscale geometry
        self.L_cc = self.geo.L_cc
        self.L = self.geo.L
        # Note: the surface area to volume ratio is defined later with the function
        # parameters. The particle size as a function of through-cell position is
        # already defined in geometric_parameters.py
        self.R_dimensional = self.geo.R_dimensional

        # Tab geometry (for pouch cells)
        self.L_tab = self.geo.L_tab
        self.Centre_y_tab = self.geo.Centre_y_tab
        self.Centre_z_tab = self.geo.Centre_z_tab
        self.A_tab = self.geo.A_tab

        # Particle properties
        self.c_max = pybamm.Parameter(
            f"Maximum concentration in {domain} electrode [mol.m-3]"
        )
        self.sigma_cc_dimensional = pybamm.Parameter(
            f"{Domain} current collector conductivity [S.m-1]"
        )
        self.epsilon_init = pybamm.FunctionParameter(
            f"{Domain} electrode porosity", {"Through-cell distance (x) [m]": x}
        )

        self.b_e = self.geo.b_e
        self.b_s = self.geo.b_s

        # Particle-size distribution parameters
        self.R_min_dim = self.geo.R_min_dim
        self.R_max_dim = self.geo.R_max_dim
        self.sd_a_dim = self.geo.sd_a_dim
        self.f_a_dist_dimensional = self.geo.f_a_dist_dimensional

        # Electrochemical reactions
        self.ne = pybamm.Parameter(f"{Domain} electrode electrons in reaction")
        self.C_dl_dimensional = pybamm.Parameter(
            f"{Domain} electrode double-layer capacity [F.m-2]"
        )

        # Intercalation kinetics
        self.mhc_lambda_dimensional = pybamm.Parameter(
            f"{Domain} electrode reorganization energy [eV]"
        )
        self.alpha_bv = pybamm.Parameter(
            f"{Domain} electrode Butler-Volmer transfer coefficient"
        )

        # Mechanical parameters
        self.nu = pybamm.Parameter(f"{Domain} electrode Poisson's ratio")
        self.E = pybamm.Parameter(f"{Domain} electrode Young's modulus [Pa]")
        self.c_0_dim = pybamm.Parameter(
            f"{Domain} electrode reference concentration for free of deformation "
            "[mol.m-3]"
        )
        self.Omega = pybamm.Parameter(
            f"{Domain} electrode partial molar volume [m3.mol-1]"
        )
        self.l_cr_0 = pybamm.Parameter(f"{Domain} electrode initial crack length [m]")
        self.w_cr = pybamm.Parameter(f"{Domain} electrode initial crack width [m]")
        self.rho_cr_dim = pybamm.Parameter(
            f"{Domain} electrode number of cracks per unit area [m-2]"
        )
        self.b_cr = pybamm.Parameter(f"{Domain} electrode Paris' law constant b")
        self.m_cr = pybamm.Parameter(f"{Domain} electrode Paris' law constant m")
        self.Eac_cr = pybamm.Parameter(
            f"{Domain} electrode activation energy for cracking rate [kJ.mol-1]"
        )
        # intermediate variables  [K*m^3/mol]
        self.theta_dim = (
            (self.Omega / main.R) * 2 * self.Omega * self.E / 9 / (1 - self.nu)
        )

        # Loss of active material parameters
        self.m_LAM = pybamm.Parameter(
            f"{Domain} electrode LAM constant exponential term"
        )
        self.beta_LAM_dimensional = pybamm.Parameter(
            f"{Domain} electrode LAM constant proportional term [s-1]"
        )
        self.stress_critical_dim = pybamm.Parameter(
            f"{Domain} electrode critical stress [Pa]"
        )
        self.beta_LAM_sei_dimensional = pybamm.Parameter(
            f"{Domain} electrode reaction-driven LAM factor [m3.mol-1]"
        )

        # utilisation parameters
        self.u_init = pybamm.Parameter(
            f"Initial {domain} electrode interface utilisation"
        )
        self.beta_utilisation_dimensional = pybamm.Parameter(
            f"{Domain} electrode current-driven interface utilisation factor [m3.mol-1]"
        )

        if self.main_param.half_cell and self.domain == "Negative":
            self.n_Li_init = pybamm.Scalar(0)
            self.U_ref = pybamm.Scalar(0)
            self.U_init_dim = pybamm.Scalar(0)
        else:
            self.epsilon_s = pybamm.FunctionParameter(
                f"{Domain} electrode active material volume fraction",
                {"Through-cell distance (x) [m]": x},
            )
            self.epsilon_inactive = 1 - self.epsilon_init - self.epsilon_s
            self.c_init = (
                pybamm.FunctionParameter(
                    f"Initial concentration in {domain} electrode [mol.m-3]",
                    {
                        "Radial distance (r) [m]": r,
                        "Through-cell distance (x) [m]": pybamm.PrimaryBroadcast(
                            x, f"{domain} particle"
                        ),
                    },
                )
                / self.c_max
            )
            c_init_av = pybamm.xyz_average(pybamm.r_average(self.c_init))
            eps_c_init_av = pybamm.xyz_average(
                self.epsilon_s * pybamm.r_average(self.c_init)
            )
            self.n_Li_init = eps_c_init_av * self.c_max * self.L * main.A_cc

            eps_s_av = pybamm.xyz_average(self.epsilon_s)
            self.elec_loading = eps_s_av * self.L * self.c_max * main.F / 3600
            self.cap_init = self.elec_loading * main.A_cc

            self.U_ref = self.U_dimensional(c_init_av, main.T_ref)
            self.U_init_dim = self.U_dimensional(c_init_av, main.T_init_dim)

    def sigma_dimensional(self, T):
        """Dimensional electrical conductivity in electrode"""
        inputs = {"Temperature [K]": T}
        return pybamm.FunctionParameter(
            f"{self.domain} electrode conductivity [S.m-1]", inputs
        )

    def D_dimensional(self, sto, T):
        """Dimensional diffusivity in particle. Note this is defined as a
        function of stochiometry"""
        inputs = {f"{self.domain} particle stoichiometry": sto, "Temperature [K]": T}
        return pybamm.FunctionParameter(
            f"{self.domain} electrode diffusivity [m2.s-1]", inputs
        )

    def j0_dimensional(self, c_e, c_s_surf, T):
        """Dimensional exchange-current density [A.m-2]"""
        inputs = {
            "Electrolyte concentration [mol.m-3]": c_e,
            f"{self.domain} particle surface concentration [mol.m-3]": c_s_surf,
            "Temperature [K]": T,
        }
        return pybamm.FunctionParameter(
            f"{self.domain} electrode exchange-current density [A.m-2]", inputs
        )

    def U_dimensional(self, sto, T):
        """Dimensional open-circuit potential [V]"""
        inputs = {f"{self.domain} particle stoichiometry": sto}
        u_ref = pybamm.FunctionParameter(f"{self.domain} electrode OCP [V]", inputs)
        # add a term to ensure that the OCP goes to infinity at 0 and -infinity at 1
        # this will not affect the OCP for most values of sto
        # see #1435
        u_ref = u_ref + 1e-6 * (1 / sto + 1 / (sto - 1))
        dudt_dim_func = self.dUdT_dimensional(sto)
        d = self.domain.lower()[0]
        dudt_dim_func.print_name = r"\frac{dU_{" + d + r"}}{dT}"
        return u_ref + (T - self.main_param.T_ref) * dudt_dim_func

    def dUdT_dimensional(self, sto):
        """
        Dimensional entropic change of the open-circuit potential [V.K-1]
        """
        inputs = {f"{self.domain} particle stoichiometry": sto}
        return pybamm.FunctionParameter(
            f"{self.domain} electrode OCP entropic change [V.K-1]", inputs
        )

    def _set_scales(self):
        """Define the scales used in the non-dimensionalisation scheme"""
        if self.domain == "Separator":
            return

        main = self.main_param
        # Microscale
        self.R_typ = self.geo.R_typ

        if main.half_cell and self.domain == "Negative":
            self.a_typ = pybamm.Scalar(1)
        elif main.options["particle shape"] == "spherical":
            self.a_typ = 3 * pybamm.xyz_average(self.epsilon_s) / self.R_typ

        # Concentration
        self.particle_concentration_scale = self.c_max

        # Scale for interfacial current density in A/m2
        if main.half_cell and self.domain == "Negative":
            # metal electrode (boundary condition between negative and separator)
            self.j_scale = main.i_typ
        else:
            # porous electrode
            self.j_scale = main.i_typ / (self.a_typ * main.L_x)

        # Reference exchange-current density
        self.j0_ref_dimensional = (
            self.j0_dimensional(main.c_e_typ, self.c_max / 2, main.T_ref) * 2
        )

        # Reaction timescales
        self.tau_r = main.F * self.c_max / (self.j0_ref_dimensional * self.a_typ)
        # Particle diffusion timescales
        self.D_typ_dim = self.D_dimensional(pybamm.Scalar(1), main.T_ref)
        self.tau_diffusion = self.R_typ ** 2 / self.D_typ_dim

    def _set_dimensionless_parameters(self):
        main = self.main_param

        if self.domain == "Separator":
            self.l = self.geo.l
            self.rho = self.therm.rho
            self.lambda_ = self.therm.lambda_
            return

        # Timescale ratios
        self.C_diff = self.tau_diffusion / main.timescale
        self.C_r = self.tau_r / main.timescale

        # Macroscale Geometry
        self.l_cc = self.geo.l_cc
        self.l = self.geo.l

        # Thermal
        self.rho_cc = self.therm.rho_cc
        self.rho = self.therm.rho
        self.lambda_cc = self.therm.lambda_cc
        self.lambda_ = self.therm.lambda_
        self.h_tab = self.therm.h_tab
        self.h_cc = self.therm.h_cc

        # Tab geometry (for pouch cells)
        self.l_tab = self.geo.l_tab
        self.centre_y_tab = self.geo.centre_y_tab
        self.centre_z_tab = self.geo.centre_z_tab

        # Microscale geometry
        self.R = self.geo.R
        self.a_R = self.a_typ * self.R_typ

        # Particle-size distribution geometry
        self.R_min = self.geo.R_min
        self.R_max = self.geo.R_max
        self.sd_a = self.geo.sd_a
        self.f_a_dist = self.geo.f_a_dist

        # Concentration ratios
        # In most cases gamma_n will be equal to 1
        self.gamma = (main.tau_discharge / main.timescale) * self.c_max / main.c_max

        # Electrode Properties
        self.sigma_cc = (
            self.sigma_cc_dimensional * main.potential_scale / main.i_typ / main.L_x
        )
        self.sigma_cc_prime = self.sigma_cc * main.delta ** 2
        self.sigma_cc_dbl_prime = self.sigma_cc_prime * main.delta

        # Electrolyte Properties
        self.beta_surf = pybamm.Scalar(0)

        # Electrochemical Reactions
        self.C_dl = (
            self.C_dl_dimensional * main.potential_scale / self.j_scale / main.timescale
        )

        # Intercalation kinetics
        self.mhc_lambda = self.mhc_lambda_dimensional / main.potential_scale_eV

        # Initial conditions
        if main.half_cell and self.domain == "Negative":
            self.U_init = pybamm.Scalar(0)
        else:
            self.U_init = (self.U_init_dim - self.U_ref) / main.potential_scale

        # Dimensionless mechanical parameters
        self.rho_cr = self.rho_cr_dim * self.l_cr_0 * self.w_cr
        self.theta = self.theta_dim * self.c_max / main.T_ref
        self.c_0 = self.c_0_dim / self.c_max
        self.beta_LAM = self.beta_LAM_dimensional * main.timescale
        # normalised typical time for one cycle
        self.stress_critical = self.stress_critical_dim / self.E
        # Reaction-driven LAM parameters
        self.beta_LAM_sei = (
            self.beta_LAM_sei_dimensional * self.a_typ * self.j_scale * main.timescale
        ) / main.F
        # Utilisation factors
        self.beta_utilisation = (
            self.beta_utilisation_dimensional
            * self.a_typ
            * self.j_scale
            * main.timescale
        ) / main.F

    def sigma(self, T):
        """Dimensionless electrode electrical conductivity"""
        main = self.main_param
        T_dim = self.main_param.Delta_T * T + self.main_param.T_ref
        return (
            self.sigma_dimensional(T_dim) * main.potential_scale / main.i_typ / main.L_x
        )

    def sigma_prime(self, T):
        """Rescaled dimensionless electrode electrical conductivity"""
        return self.sigma(T) * self.main_param.delta

    def D(self, c_s, T):
        """Dimensionless particle diffusivity"""
        sto = c_s
        T_dim = self.main_param.Delta_T * T + self.main_param.T_ref
        return self.D_dimensional(sto, T_dim) / self.D_typ_dim

    def j0(self, c_e, c_s_surf, T):
        """Dimensionless exchange-current density"""
        c_e_dim = c_e * self.main_param.c_e_typ
        c_s_surf_dim = c_s_surf * self.c_max
        T_dim = self.main_param.Delta_T * T + self.main_param.T_ref

        return (
            self.j0_dimensional(c_e_dim, c_s_surf_dim, T_dim) / self.j0_ref_dimensional
        )

    def U(self, c_s, T):
        """Dimensionless open-circuit potential in the electrode"""
        main = self.main_param
        sto = c_s
        T_dim = self.main_param.Delta_T * T + self.main_param.T_ref
        return (self.U_dimensional(sto, T_dim) - self.U_ref) / main.potential_scale

    def dUdT(self, c_s):
        """Dimensionless entropic change in open-circuit potential"""
        main = self.main_param
        sto = c_s
        return self.dUdT_dimensional(sto) * main.Delta_T / main.potential_scale

    def t_change(self, sto):
        """
        Dimensionless volume change for the electrode;
        sto should be R-averaged
        """
        return pybamm.FunctionParameter(
            f"{self.domain} electrode volume change", {"Particle stoichiometry": sto}
        )

    def k_cr(self, T):
        """
        Dimensionless cracking rate for the electrode;
        """
        T_dim = self.main_param.Delta_T * T + self.main_param.T_ref
        delta_k_cr = self.E ** self.m_cr * self.l_cr_0 ** (self.m_cr / 2 - 1)
        return (
            pybamm.FunctionParameter(
                f"{self.domain} electrode cracking rate", {"Temperature [K]": T_dim}
            )
            * delta_k_cr
        )



#
# Write a symbol to Julia
#
import pybamm

import numpy as np
import scipy.sparse
from collections import OrderedDict

import numbers


def id_to_julia_variable(symbol_id, prefix):
    """
    This function defines the format for the julia variable names used in find_symbols
    and to_julia. Variable names are based on a nodes' id to make them unique
    """
    var_format = prefix + "_{:05d}"
    # Need to replace "-" character to make them valid julia variable names
    return var_format.format(symbol_id).replace("-", "m")


def is_constant_and_can_evaluate(symbol):
    """
    Returns True if symbol is constant and evaluation does not raise any errors.
    Returns False otherwise.
    An example of a constant symbol that cannot be "evaluated" is PrimaryBroadcast(0).
    """
    if symbol.is_constant():
        try:
            symbol.evaluate()
            return True
        except NotImplementedError:
            return False
    else:
        return False


def find_symbols(
    symbol,
    constant_symbols,
    variable_symbols,
    variable_symbol_sizes,
    round_constants=True,
):
    """
    This function converts an expression tree to a dictionary of node id's and strings
    specifying valid julia code to calculate that nodes value, given y and t.

    The function distinguishes between nodes that represent constant nodes in the tree
    (e.g. a pybamm.Matrix), and those that are variable (e.g. subtrees that contain
    pybamm.StateVector). The former are put in `constant_symbols`, the latter in
    `variable_symbols`

    Note that it is important that the arguments `constant_symbols` and
    `variable_symbols` be and *ordered* dict, since the final ordering of the code lines
    are important for the calculations. A dict is specified rather than a list so that
    identical subtrees (which give identical id's) are not recalculated in the code

    Parameters
    ----------
    symbol : :class:`pybamm.Symbol`
        The symbol or expression tree to convert

    constant_symbol : collections.OrderedDict
        The output dictionary of constant symbol ids to lines of code

    variable_symbol : collections.OrderedDict
        The output dictionary of variable (with y or t) symbol ids to lines of code

    variable_symbol_sizes : collections.OrderedDict
        The output dictionary of variable (with y or t) symbol ids to size of that
        variable, for caching

    """
    # ignore broadcasts for now
    if isinstance(symbol, pybamm.Broadcast):
        symbol = symbol.child
    if is_constant_and_can_evaluate(symbol):
        value = symbol.evaluate()
        if round_constants:
            value = np.round(value, decimals=11)
        if not isinstance(value, numbers.Number):
            if scipy.sparse.issparse(value):
                # Create Julia SparseArray
                row, col, data = scipy.sparse.find(value)
                if round_constants:
                    data = np.round(data, decimals=11)
                m, n = value.shape
                # Set print options large enough to avoid ellipsis
                # at least as big is len(row) = len(col) = len(data)
                np.set_printoptions(
                    threshold=max(np.get_printoptions()["threshold"], len(row) + 10)
                )
                # increase precision for printing
                np.set_printoptions(precision=20)
                # add 1 to correct for 1-indexing in Julia
                # use array2string so that commas are included
                constant_symbols[symbol.id] = "sparse({}, {}, {}, {}, {})".format(
                    np.array2string(row + 1, separator=","),
                    np.array2string(col + 1, separator=","),
                    np.array2string(data, separator=","),
                    m,
                    n,
                )
            elif value.shape == (1, 1):
                # Extract value if array has only one entry
                constant_symbols[symbol.id] = value[0, 0]
                variable_symbol_sizes[symbol.id] = 1
            elif value.shape[1] == 1:
                # Set print options large enough to avoid ellipsis
                # at least as big as len(row) = len(col) = len(data)
                np.set_printoptions(
                    threshold=max(
                        np.get_printoptions()["threshold"], value.shape[0] + 10
                    )
                )
                # Flatten a 1D array
                constant_symbols[symbol.id] = np.array2string(
                    value.flatten(), separator=","
                )
                variable_symbol_sizes[symbol.id] = symbol.shape[0]
            else:
                constant_symbols[symbol.id] = value
                # No need to save the size as it will not need to be used
        return

    # process children recursively
    for child in symbol.children:
        find_symbols(
            child,
            constant_symbols,
            variable_symbols,
            variable_symbol_sizes,
            round_constants=round_constants,
        )

    # calculate the variable names that will hold the result of calculating the
    # children variables
    children_vars = []
    for child in symbol.children:
        if isinstance(child, pybamm.Broadcast):
            child = child.child
        if is_constant_and_can_evaluate(child):
            child_eval = child.evaluate()
            if isinstance(child_eval, numbers.Number):
                children_vars.append(str(child_eval))
            else:
                children_vars.append(id_to_julia_variable(child.id, "const"))
        else:
            children_vars.append(id_to_julia_variable(child.id, "cache"))

    if isinstance(symbol, pybamm.BinaryOperator):
        # TODO: we can pass through a dummy y and t to get the type and then hardcode
        # the right line, avoiding these checks
        if isinstance(symbol, pybamm.MatrixMultiplication):
            symbol_str = "{0} @ {1}".format(children_vars[0], children_vars[1])
        elif isinstance(symbol, pybamm.Inner):
            symbol_str = "{0} * {1}".format(children_vars[0], children_vars[1])
        elif isinstance(symbol, pybamm.Minimum):
            symbol_str = "min({},{})".format(children_vars[0], children_vars[1])
        elif isinstance(symbol, pybamm.Maximum):
            symbol_str = "max({},{})".format(children_vars[0], children_vars[1])
        elif isinstance(symbol, pybamm.Power):
            # julia uses ^ instead of ** for power
            # include dot for elementwise operations
            symbol_str = children_vars[0] + " .^ " + children_vars[1]
        else:
            # all other operations use the same symbol
            symbol_str = children_vars[0] + " " + symbol.name + " " + children_vars[1]

    elif isinstance(symbol, pybamm.UnaryOperator):
        # Index has a different syntax than other univariate operations
        if isinstance(symbol, pybamm.Index):
            # Because of how julia indexing works, add 1 to the start, but not to the
            # stop
            symbol_str = "{}[{}:{}]".format(
                children_vars[0], symbol.slice.start + 1, symbol.slice.stop
            )
        elif isinstance(symbol, pybamm.Gradient):
            symbol_str = "grad_{}({})".format(tuple(symbol.domain), children_vars[0])
        elif isinstance(symbol, pybamm.Divergence):
            symbol_str = "div_{}({})".format(tuple(symbol.domain), children_vars[0])
        elif isinstance(symbol, pybamm.Broadcast):
            # ignore broadcasts for now
            symbol_str = children_vars[0]
        elif isinstance(symbol, pybamm.BoundaryValue):
            symbol_str = "boundary_value_{}({})".format(symbol.side, children_vars[0])
        else:
            symbol_str = symbol.name + children_vars[0]

    elif isinstance(symbol, pybamm.Function):
        # write functions directly
        symbol_str = "{}({})".format(symbol.julia_name, ", ".join(children_vars))

    elif isinstance(symbol, (pybamm.Variable, pybamm.ConcatenationVariable)):
        # No need to do anything if a Variable is found
        return

    elif isinstance(symbol, pybamm.Concatenation):
        if isinstance(symbol, (pybamm.NumpyConcatenation, pybamm.SparseStack)):
            # return a list of the children variables, which will be converted to a
            # line by line assignment
            # return this as a string so that other functionality still works
            # also save sizes
            symbol_str = "["
            for child in children_vars:
                child_id = child[6:].replace("m", "-")
                size = variable_symbol_sizes[int(child_id)]
                symbol_str += "{}::{}, ".format(size, child)
            symbol_str = symbol_str[:-2] + "]"

        # DomainConcatenation specifies a particular ordering for the concatenation,
        # which we must follow
        elif isinstance(symbol, pybamm.DomainConcatenation):
            if symbol.secondary_dimensions_npts == 1:
                all_child_vectors = children_vars
                all_child_sizes = [
                    variable_symbol_sizes[int(child[6:].replace("m", "-"))]
                    for child in children_vars
                ]
            else:
                slice_starts = []
                all_child_vectors = []
                all_child_sizes = []
                for i in range(symbol.secondary_dimensions_npts):
                    child_vectors = []
                    child_sizes = []
                    for child_var, slices in zip(
                        children_vars, symbol._children_slices
                    ):
                        for child_dom, child_slice in slices.items():
                            slice_starts.append(symbol._slices[child_dom][i].start)
                            # add 1 to slice start to account for julia indexing
                            child_vectors.append(
                                "@view {}[{}:{}]".format(
                                    child_var,
                                    child_slice[i].start + 1,
                                    child_slice[i].stop,
                                )
                            )
                            child_sizes.append(
                                child_slice[i].stop - child_slice[i].start
                            )
                    all_child_vectors.extend(
                        [v for _, v in sorted(zip(slice_starts, child_vectors))]
                    )
                    all_child_sizes.extend(
                        [v for _, v in sorted(zip(slice_starts, child_sizes))]
                    )
            # return a list of the children variables, which will be converted to a
            # line by line assignment
            # return this as a string so that other functionality still works
            # also save sizes
            symbol_str = "["
            for child, size in zip(all_child_vectors, all_child_sizes):
                symbol_str += "{}::{}, ".format(size, child)
            symbol_str = symbol_str[:-2] + "]"

        else:
            # A regular Concatenation for the MTK model
            # We will define the concatenation function separately
            symbol_str = "concatenation(" + ", ".join(children_vars) + ")"

    # Note: we assume that y is being passed as a column vector
    elif isinstance(symbol, pybamm.StateVectorBase):
        if isinstance(symbol, pybamm.StateVector):
            name = "@view y"
        elif isinstance(symbol, pybamm.StateVectorDot):
            name = "@view dy"
        indices = np.argwhere(symbol.evaluation_array).reshape(-1).astype(np.int32)
        # add 1 since julia uses 1-indexing
        indices += 1
        if len(indices) == 1:
            symbol_str = "{}[{}]".format(name, indices[0])
        else:
            # julia does include the final value
            symbol_str = "{}[{}:{}]".format(name, indices[0], indices[-1])

    elif isinstance(symbol, pybamm.Time):
        symbol_str = "t"

    elif isinstance(symbol, pybamm.InputParameter):
        symbol_str = "inputs['{}']".format(symbol.name)

    elif isinstance(symbol, pybamm.SpatialVariable):
        symbol_str = symbol.name

    elif isinstance(symbol, pybamm.FunctionParameter):
        symbol_str = "{}({})".format(symbol.name, ", ".join(children_vars))

    else:
        raise NotImplementedError(
            "Conversion to Julia not implemented for a symbol of type '{}'".format(
                type(symbol)
            )
        )

    variable_symbols[symbol.id] = symbol_str

    # Save the size of the symbol
    try:
        if symbol.shape == ():
            variable_symbol_sizes[symbol.id] = 1
        else:
            variable_symbol_sizes[symbol.id] = symbol.shape[0]
    except NotImplementedError:
        pass


def to_julia(symbol, round_constants=True):
    """
    This function converts an expression tree into a dict of constant input values, and
    valid julia code that acts like the tree's :func:`pybamm.Symbol.evaluate` function

    Parameters
    ----------
    symbol : :class:`pybamm.Symbol`
        The symbol to convert to julia code

    Returns
    -------
    constant_values : collections.OrderedDict
        dict mapping node id to a constant value. Represents all the constant nodes in
        the expression tree
    str
        valid julia code that will evaluate all the variable nodes in the tree.

    """

    constant_values = OrderedDict()
    variable_symbols = OrderedDict()
    variable_symbol_sizes = OrderedDict()
    find_symbols(
        symbol,
        constant_values,
        variable_symbols,
        variable_symbol_sizes,
        round_constants=round_constants,
    )

    return constant_values, variable_symbols, variable_symbol_sizes


def get_julia_function(
    symbol,
    funcname="f",
    input_parameter_order=None,
    len_rhs=None,
    preallocate=True,
    round_constants=True,
):
    """
    Converts a pybamm expression tree into pure julia code that will calculate the
    result of calling `evaluate(t, y)` on the given expression tree.

    Parameters
    ----------
    symbol : :class:`pybamm.Symbol`
        The symbol to convert to julia code
    funcname : str, optional
        The name to give to the function (default 'f')
    input_parameter_order : list, optional
        List of input parameter names. Defines the order in which the input parameters
        are extracted from 'p' in the julia function that is created
    len_rhs : int, optional
        The number of ODEs in the discretized differential equations. This also
        determines whether the model has any algebraic equations: if None (default),
        the model is assume to have no algebraic parts and ``julia_str`` is compatible
        with an ODE solver. If not None, ``julia_str`` is compatible with a DAE solver
    preallocate : bool, optional
        Whether to write the function in a way that preallocates memory for the output.
        Default is True, which is faster. Must be False for the function to be
        modelingtoolkitized.

    Returns
    -------
    julia_str : str
        String of julia code, to be evaluated by ``julia.Main.eval``

    """
    if len_rhs is None:
        typ = "ode"
    else:
        typ = "dae"
        # Take away dy from the differential states
        # we will return a function of the form
        # out[] = .. - dy[] for the differential states
        # out[] = .. for the algebraic states
        symbol_minus_dy = []
        end = 0
        for child in symbol.orphans:
            start = end
            end += child.size
            if end <= len_rhs:
                symbol_minus_dy.append(child - pybamm.StateVectorDot(slice(start, end)))
            else:
                symbol_minus_dy.append(child)
        symbol = pybamm.numpy_concatenation(*symbol_minus_dy)
    constants, var_symbols, var_symbol_sizes = to_julia(
        symbol, round_constants=round_constants
    )

    # extract constants in generated function
    const_and_cache_str = "cs = (\n"
    shorter_const_names = {}
    for i_const, (symbol_id, const_value) in enumerate(constants.items()):
        const_name = id_to_julia_variable(symbol_id, "const")
        const_name_short = "const_{}".format(i_const)
        const_and_cache_str += "   {} = {},\n".format(const_name_short, const_value)
        shorter_const_names[const_name] = const_name_short

    # Pop (get and remove) items from the dictionary of symbols one by one
    # If they are simple operations (@view, +, -, *, /), replace all future
    # occurences instead of assigning them. This "inlining" speeds up the computation
    inlineable_symbols = ["@view", "+", "-", "*", "/"]
    var_str = ""
    input_parameters = {}
    while var_symbols:
        var_symbol_id, symbol_line = var_symbols.popitem(last=False)
        julia_var = id_to_julia_variable(var_symbol_id, "cache")
        # Look for lists in the variable symbols. These correpsond to concatenations, so
        # assign the children to the right parts of the vector
        if symbol_line[0] == "[" and symbol_line[-1] == "]":
            # convert to actual list
            symbol_line = symbol_line[1:-1].split(", ")
            start = 0
            if preallocate is True or var_symbol_id == symbol.id:
                for child_size_and_name in symbol_line:
                    child_size, child_name = child_size_and_name.split("::")
                    end = start + int(child_size)
                    # add 1 to start to account for julia 1-indexing
                    var_str += "@. {}[{}:{}] = {}\n".format(
                        julia_var, start + 1, end, child_name
                    )
                    start = end
            else:
                concat_str = "{} = vcat(".format(julia_var)
                for i, child_size_and_name in enumerate(symbol_line):
                    child_size, child_name = child_size_and_name.split("::")
                    var_str += "x{} = @. {}\n".format(i + 1, child_name)
                    concat_str += "x{}, ".format(i + 1)
                var_str += concat_str[:-2] + ")\n"
        # use mul! for matrix multiplications (requires LinearAlgebra library)
        elif " @ " in symbol_line:
            if preallocate is False:
                symbol_line = symbol_line.replace(" @ ", " * ")
                var_str += "{} = {}\n".format(julia_var, symbol_line)
            else:
                symbol_line = symbol_line.replace(" @ ", ", ")
                var_str += "mul!({}, {})\n".format(julia_var, symbol_line)
        # find input parameters
        elif symbol_line.startswith("inputs"):
            input_parameters[julia_var] = symbol_line[8:-2]
        elif "minimum" in symbol_line or "maximum" in symbol_line:
            var_str += "{} .= {}\n".format(julia_var, symbol_line)
        else:
            # don't replace the matrix multiplication cases (which will be
            # turned into a mul!), since it is faster to assign to a cache array
            # first in that case
            # e.g. mul!(cs.cache_1, cs.cache_2, cs.cache_3)
            # unless it is a @view in which case we don't
            # need to cache
            # e.g. mul!(cs.cache_1, cs.cache_2, @view y[1:10])
            # also don't replace the minimum() or maximum() cases as we can't
            # broadcast them
            any_matmul_min_max = any(
                julia_var in next_symbol_line
                and (
                    any(
                        x in next_symbol_line
                        for x in [" @ ", "mul!", "minimum", "maximum"]
                    )
                    and not symbol_line.startswith("@view")
                )
                for next_symbol_line in var_symbols.values()
            )
            # inline operation if it can be inlined
            if (
                any(x in symbol_line for x in inlineable_symbols) or symbol_line == "t"
            ) and not any_matmul_min_max:
                found_replacement = False
                # replace all other occurrences of the variable
                # in the dictionary with the symbol line
                for next_var_id, next_symbol_line in var_symbols.items():
                    if julia_var in next_symbol_line:
                        if symbol_line == "t":
                            # no brackets needed
                            var_symbols[next_var_id] = next_symbol_line.replace(
                                julia_var, symbol_line
                            )
                        else:
                            # add brackets so that the order of operations is maintained
                            var_symbols[next_var_id] = next_symbol_line.replace(
                                julia_var, "({})".format(symbol_line)
                            )
                        found_replacement = True
                if not found_replacement:
                    var_str += "@. {} = {}\n".format(julia_var, symbol_line)

            # otherwise assign
            else:
                var_str += "@. {} = {}\n".format(julia_var, symbol_line)
    # Replace all input parameter names
    for input_parameter_id, input_parameter_name in input_parameters.items():
        var_str = var_str.replace(input_parameter_id, input_parameter_name)

    # indent code
    var_str = "   " + var_str
    var_str = var_str.replace("\n", "\n   ")

    # add the cache variables to the cache NamedTuple
    i_cache = 0
    for var_symbol_id, var_symbol_size in var_symbol_sizes.items():
        # Skip caching the result variable since this is provided as dy
        # Also skip caching the result variable if it doesn't appear in the var_str,
        # since it has been inlined and does not need to be assigned to
        julia_var = id_to_julia_variable(var_symbol_id, "cache")
        if var_symbol_id != symbol.id and julia_var in var_str:
            julia_var_short = "cache_{}".format(i_cache)
            var_str = var_str.replace(julia_var, julia_var_short)
            i_cache += 1
            if preallocate is True:
                const_and_cache_str += "   {} = zeros({}),\n".format(
                    julia_var_short, var_symbol_size
                )
            else:
                # Cache variables have not been preallocated
                var_str = var_str.replace(
                    "@. {} = ".format(julia_var_short),
                    "{} = @. ".format(julia_var_short),
                )

    # Shorten the name of the constants from id to const_0, const_1, etc.
    for long, short in shorter_const_names.items():
        var_str = var_str.replace(long, "cs." + short)

    # close the constants and cache string
    const_and_cache_str += ")\n"

    # remove the constant and cache sring if it is empty
    const_and_cache_str = const_and_cache_str.replace("cs = (\n)\n", "")

    # calculate the final variable that will output the result
    if symbol.is_constant():
        result_var = id_to_julia_variable(symbol.id, "const")
        if result_var in shorter_const_names:
            result_var = shorter_const_names[result_var]
        result_value = symbol.evaluate()
        if isinstance(result_value, numbers.Number):
            var_str = var_str + "\n   dy .= " + str(result_value) + "\n"
        else:
            var_str = var_str + "\n   dy .= cs." + result_var + "\n"
    else:
        result_var = id_to_julia_variable(symbol.id, "cache")
        if typ == "ode":
            out = "dy"
        elif typ == "dae":
            out = "out"
        # replace "cache_123 = ..." with "dy .= ..." (ensure we allocate to the
        # variable that was passed in)
        var_str = var_str.replace(f"   {result_var} =", f"   {out} .=")
        # catch other cases for dy
        var_str = var_str.replace(result_var, out)

    # add "cs." to cache names
    if preallocate is True:
        var_str = var_str.replace("cache", "cs.cache")

    # line that extracts the input parameters in the right order
    if input_parameter_order is None:
        input_parameter_extraction = ""
    elif len(input_parameter_order) == 1:
        # extract the single parameter
        input_parameter_extraction = "   " + input_parameter_order[0] + " = p[1]\n"
    else:
        # extract all parameters
        input_parameter_extraction = "   " + ", ".join(input_parameter_order) + " = p\n"

    if preallocate is False or const_and_cache_str == "":
        func_def = f"{funcname}!"
    else:
        func_def = f"{funcname}_with_consts!"

    # add function def
    if typ == "ode":
        function_def = f"\nfunction {func_def}(dy, y, p, t)\n"
    elif typ == "dae":
        function_def = f"\nfunction {func_def}(out, dy, y, p, t)\n"
    julia_str = (
        "begin\n"
        + const_and_cache_str
        + function_def
        + input_parameter_extraction
        + var_str
    )

    # close the function, with a 'nothing' to avoid allocations
    julia_str += "nothing\nend\n\n"
    julia_str = julia_str.replace("\n   \n", "\n")

    if not (preallocate is False or const_and_cache_str == ""):
        # Use a let block for the cached variables
        # open the let block
        julia_str = julia_str.replace("cs = (", f"{funcname}! = let cs = (")
        # close the let block
        julia_str += "end\n"

    # close the "begin"
    julia_str += "end"

    return julia_str


def convert_var_and_eqn_to_str(var, eqn, all_constants_str, all_variables_str, typ):
    """
    Converts a variable and its equation to a julia string

    Parameters
    ----------
    var : :class:`pybamm.Symbol`
        The variable (key in the dictionary of rhs/algebraic/initial conditions)
    eqn : :class:`pybamm.Symbol`
        The equation (value in the dictionary of rhs/algebraic/initial conditions)
    all_constants_str : str
        String containing all the constants defined so far
    all_variables_str : str
        String containing all the variables defined so far
    typ : str
        The type of the variable/equation pair being converted ("equation", "initial
        condition", or "boundary condition")

    Returns
    -------
    all_constants_str : str
        Updated string of all constants
    all_variables_str : str
        Updated string of all variables
    eqn_str : str
        The string describing the final equation result, perhaps as a function of some
        variables and/or constants in all_constants_str and all_variables_str

    """
    if isinstance(eqn, pybamm.Broadcast):
        # ignore broadcasts for now
        eqn = eqn.child

    var_symbols = to_julia(eqn)[1]

    # var_str = ""
    # for symbol_id, symbol_line in var_symbols.items():
    #     var_str += f"{id_to_julia_variable(symbol_id)} = {symbol_line}\n"
    # Pop (get and remove) items from the dictionary of symbols one by one
    # If they are simple operations (+, -, *, /), replace all future
    # occurences instead of assigning them.
    inlineable_symbols = [" + ", " - ", " * ", " / "]
    var_str = ""
    while var_symbols:
        var_symbol_id, symbol_line = var_symbols.popitem(last=False)
        julia_var = id_to_julia_variable(var_symbol_id, "cache")
        # inline operation if it can be inlined
        if "concatenation" not in symbol_line:
            found_replacement = False
            # replace all other occurrences of the variable
            # in the dictionary with the symbol line
            for next_var_id, next_symbol_line in var_symbols.items():
                if (
                    symbol_line == "t"
                    or " " not in symbol_line
                    or symbol_line.startswith("grad")
                    or not any(x in next_symbol_line for x in inlineable_symbols)
                ):
                    # cases that don't need brackets
                    var_symbols[next_var_id] = next_symbol_line.replace(
                        julia_var, symbol_line
                    )
                elif next_symbol_line.startswith("concatenation"):
                    var_symbols[next_var_id] = next_symbol_line.replace(
                        julia_var, f"\n   {symbol_line}\n"
                    )
                else:
                    # add brackets so that the order of operations is maintained
                    var_symbols[next_var_id] = next_symbol_line.replace(
                        julia_var, "({})".format(symbol_line)
                    )
                found_replacement = True
            if not found_replacement:
                var_str += "{} = {}\n".format(julia_var, symbol_line)

        # otherwise assign
        else:
            var_str += "{} = {}\n".format(julia_var, symbol_line)

    # If we have created a concatenation we need to define it
    # Hardcoded to the negative electrode, separator, positive electrode case for now
    if "concatenation" in var_str and "function concatenation" not in all_variables_str:
        concatenation_def = (
            "\nfunction concatenation(n, s, p)\n"
            + "   # A concatenation in the electrolyte domain\n"
            + "   IfElse.ifelse(\n"
            + "      x < neg_width, n, IfElse.ifelse(\n"
            + "         x < neg_plus_sep_width, s, p\n"
            + "      )\n"
            + "   )\n"
            + "end\n"
        )
    else:
        concatenation_def = ""

    # Define the FunctionParameter objects that have not yet been defined
    function_defs = ""
    for x in eqn.pre_order():
        if (
            isinstance(x, pybamm.FunctionParameter)
            and f"function {x.name}" not in all_variables_str
            and typ == "equation"
        ):
            function_def = (
                f"\nfunction {x.name}("
                + ", ".join(x.arg_names)
                + ")\n"
                + "   {}\n".format(str(x.callable).replace("**", "^"))
                + "end\n"
            )
            function_defs += function_def

    if concatenation_def + function_defs != "":
        function_defs += "\n"

    var_str = concatenation_def + function_defs + var_str

    # add a comment labeling the equation, and the equation itself
    if var_str == "":
        all_variables_str += ""
    else:
        all_variables_str += f"# '{var.name}' {typ}\n" + var_str + "\n"

    # calculate the final variable that will output the result
    if eqn.is_constant():
        result_var = id_to_julia_variable(eqn.id, "const")
    else:
        result_var = id_to_julia_variable(eqn.id, "cache")
    if is_constant_and_can_evaluate(eqn):
        result_value = eqn.evaluate()
    else:
        result_value = None

    # define the variable that goes into the equation
    if eqn.is_constant() and isinstance(result_value, numbers.Number):
        eqn_str = str(result_value)
    else:
        eqn_str = result_var

    return all_constants_str, all_variables_str, eqn_str


def get_julia_mtk_model(model, geometry=None, tspan=None):
    """
    Converts a pybamm model into a Julia ModelingToolkit model

    Parameters
    ----------
    model : :class:`pybamm.BaseModel`
        The model to be converted
    geometry : dict, optional
        Dictionary defining the geometry. Must be provided if the model is a PDE model
    tspan : array-like, optional
        Time for which to solve the model. Must be provided if the model is a PDE model

    Returns
    -------
    mtk_str : str
        String of julia code representing a model in MTK,
        to be evaluated by ``julia.Main.eval``
    """
    # Extract variables
    variables = {**model.rhs, **model.algebraic}.keys()
    variable_id_to_print_name = {}
    for i, var in enumerate(variables):
        if var.print_name is not None:
            print_name = var._raw_print_name
        else:
            print_name = f"u{i+1}"
        variable_id_to_print_name[var.id] = print_name
        if isinstance(var, pybamm.ConcatenationVariable):
            for child in var.children:
                variable_id_to_print_name[child.id] = print_name

    # Extract domain and auxiliary domains
    all_domains = set(
        [tuple(dom) for var in variables for dom in var.domains.values() if dom != []]
    )
    is_pde = bool(all_domains)

    # Check geometry and tspan have been provided if a PDE
    if is_pde:
        if geometry is None:
            raise ValueError("must provide geometry if the model is a PDE model")
        if tspan is None:
            raise ValueError("must provide tspan if the model is a PDE model")

    # Read domain names
    domain_name_to_symbol = {}
    long_domain_symbol_to_short = {}
    for dom in all_domains:
        # Read domain name from geometry
        domain_symbol = list(geometry[dom[0]].keys())[0]
        if len(dom) > 1:
            domain_symbol = domain_symbol[0]
            # For multi-domain variables keep only the first letter of the domain
            domain_name_to_symbol[tuple(dom)] = domain_symbol
            # Record which domain symbols we shortened
            for d in dom:
                long = list(geometry[d].keys())[0]
                long_domain_symbol_to_short[long] = domain_symbol
        else:
            # Otherwise keep the whole domain
            domain_name_to_symbol[tuple(dom)] = domain_symbol

    # Read domain limits
    domain_name_to_limits = {(): None}
    for dom in all_domains:
        limits = list(geometry[dom[0]].values())[0].values()
        if len(limits) > 1:
            lower_limit, _ = list(geometry[dom[0]].values())[0].values()
            _, upper_limit = list(geometry[dom[-1]].values())[0].values()
            domain_name_to_limits[tuple(dom)] = (
                lower_limit.evaluate(),
                upper_limit.evaluate(),
            )
        else:
            # Don't record limits for variables that have "limits" of length 1 i.e.
            # a zero-dimensional domain
            domain_name_to_limits[tuple(dom)] = None

    # Define independent variables for each variable
    var_to_ind_vars = {}
    var_to_ind_vars_left_boundary = {}
    var_to_ind_vars_right_boundary = {}
    for var in variables:
        if var.domain in [[], ["current collector"]]:
            var_to_ind_vars[var.id] = "(t)"
        else:
            # all independent variables e.g. (t, x) or (t, rn, xn)
            domain_symbols = ", ".join(
                domain_name_to_symbol[tuple(dom)]
                for dom in var.domains.values()
                if domain_name_to_limits[tuple(dom)] is not None
            )
            var_to_ind_vars[var.id] = f"(t, {domain_symbols})"
            if isinstance(var, pybamm.ConcatenationVariable):
                for child in var.children:
                    var_to_ind_vars[child.id] = f"(t, {domain_symbols})"
            aux_domain_symbols = ", ".join(
                domain_name_to_symbol[tuple(dom)]
                for level, dom in var.domains.items()
                if level != "primary" and domain_name_to_limits[tuple(dom)] is not None
            )
            if aux_domain_symbols != "":
                aux_domain_symbols = ", " + aux_domain_symbols

            limits = domain_name_to_limits[tuple(var.domain)]
            # left bc e.g. (t, 0) or (t, 0, xn)
            var_to_ind_vars_left_boundary[
                var.id
            ] = f"(t, {limits[0]}{aux_domain_symbols})"
            # right bc e.g. (t, 1) or (t, 1, xn)
            var_to_ind_vars_right_boundary[
                var.id
            ] = f"(t, {limits[1]}{aux_domain_symbols})"

    mtk_str = "begin\n"
    # Define parameters (including independent variables)
    # Makes a line of the form '@parameters t x1 x2 x3 a b c d'
    ind_vars = ["t"] + [
        sym
        for dom, sym in domain_name_to_symbol.items()
        if domain_name_to_limits[dom] is not None
    ]
    for domain_name, domain_symbol in domain_name_to_symbol.items():
        if domain_name_to_limits[domain_name] is not None:
            mtk_str += f"# {domain_name} -> {domain_symbol}\n"
    mtk_str += "@parameters " + " ".join(ind_vars)
    if len(model.input_parameters) > 0:
        mtk_str += "\n# Input parameters\n@parameters"
        for param in model.input_parameters:
            mtk_str += f" {param.name}"
    mtk_str += "\n"

    # Add a comment with the variable names
    for var in variables:
        mtk_str += f"# '{var.name}' -> {variable_id_to_print_name[var.id]}\n"
    # Makes a line of the form '@variables u1(t) u2(t)'
    dep_vars = []
    mtk_str += "@variables"
    for var in variables:
        mtk_str += f" {variable_id_to_print_name[var.id]}(..)"
        dep_var = variable_id_to_print_name[var.id] + var_to_ind_vars[var.id]
        dep_vars.append(dep_var)
    mtk_str += "\n"

    # Define derivatives
    for domain_symbol in ind_vars:
        mtk_str += f"D{domain_symbol} = Differential({domain_symbol})\n"
    mtk_str += "\n"

    # Define equations
    all_eqns_str = ""
    all_constants_str = ""
    all_julia_str = ""
    for var, eqn in {**model.rhs, **model.algebraic}.items():
        all_constants_str, all_julia_str, eqn_str = convert_var_and_eqn_to_str(
            var, eqn, all_constants_str, all_julia_str, "equation"
        )

        if var in model.rhs:
            all_eqns_str += (
                f"   Dt({variable_id_to_print_name[var.id]}{var_to_ind_vars[var.id]}) "
                + f"~ {eqn_str},\n"
            )
        elif var in model.algebraic:
            all_eqns_str += f"   0 ~ {eqn_str},\n"

    # Replace any long domain symbols with the short version
    # e.g. "xn" gets replaced with "x"
    for long, short in long_domain_symbol_to_short.items():
        # we need to add a space to avoid accidentally replacing 'exp' with 'ex'
        all_julia_str = all_julia_str.replace(" " + long, " " + short)

    # Replace variables in the julia strings that correspond to pybamm variables with
    # their julia equivalent
    for var_id, julia_id in variable_id_to_print_name.items():
        # e.g. boundary_value_right(cache_123456789) gets replaced with u1(t, 1)
        cache_var_id = id_to_julia_variable(var_id, "cache")
        if f"boundary_value_right({cache_var_id})" in all_julia_str:
            all_julia_str = all_julia_str.replace(
                f"boundary_value_right({cache_var_id})",
                julia_id + var_to_ind_vars_right_boundary[var_id],
            )
        # e.g. cache_123456789 gets replaced with u1(t, x)
        all_julia_str = all_julia_str.replace(
            cache_var_id, julia_id + var_to_ind_vars[var_id]
        )

    # Replace independent variables (domain names) in julia strings with the
    # corresponding symbol
    for domain_name, domain_symbol in domain_name_to_symbol.items():
        all_julia_str = all_julia_str.replace(
            f"grad_{domain_name}", f"D{domain_symbol}"
        )
        # Different divergence depending on the coordinate system
        coord_sys = getattr(pybamm.standard_spatial_vars, domain_symbol).coord_sys
        if coord_sys == "cartesian":
            all_julia_str = all_julia_str.replace(
                f"div_{domain_name}", f"D{domain_symbol}"
            )
        elif coord_sys == "spherical polar":
            all_julia_str = all_julia_str.replace(
                f"div_{domain_name}(",
                f"1 / {domain_symbol}^2 * D{domain_symbol}({domain_symbol}^2 * ",
            )

    # Replace the thicknesses in the concatenation with the actual thickness from the
    # geometry
    if "neg_width" in all_julia_str or "neg_plus_sep_width" in all_julia_str:
        var = pybamm.standard_spatial_vars
        x_n = geometry["negative electrode"]["x_n"]["max"].evaluate()
        x_s = geometry["separator"]["x_s"]["max"].evaluate()
        all_julia_str = all_julia_str.replace("neg_width", str(x_n))
        all_julia_str = all_julia_str.replace("neg_plus_sep_width", str(x_s))

    # Update the MTK string
    mtk_str += all_constants_str + all_julia_str + "\n" + f"eqs = [\n{all_eqns_str}]\n"

    ####################################################################################
    # Initial and boundary conditions
    ####################################################################################
    # Initial conditions
    all_ic_bc_str = "   # initial conditions\n"
    all_ic_bc_constants_str = ""
    all_ic_bc_julia_str = ""
    for var, eqn in model.initial_conditions.items():
        (
            all_ic_bc_constants_str,
            all_ic_bc_julia_str,
            eqn_str,
        ) = convert_var_and_eqn_to_str(
            var, eqn, all_ic_bc_constants_str, all_ic_bc_julia_str, "initial condition"
        )

        if not is_pde:
            all_ic_bc_str += (
                f"   {variable_id_to_print_name[var.id]}(t) => {eqn_str},\n"
            )
        else:
            if var.domain == []:
                doms = ""
            else:
                doms = ", " + domain_name_to_symbol[tuple(var.domain)]

            all_ic_bc_str += (
                f"   {variable_id_to_print_name[var.id]}(0{doms}) ~ {eqn_str},\n"
            )
    # Boundary conditions
    if is_pde:
        all_ic_bc_str += "   # boundary conditions\n"
        for var, eqn_side in model.boundary_conditions.items():
            if isinstance(var, (pybamm.Variable, pybamm.ConcatenationVariable)):
                for side, (eqn, typ) in eqn_side.items():
                    (
                        all_ic_bc_constants_str,
                        all_ic_bc_julia_str,
                        eqn_str,
                    ) = convert_var_and_eqn_to_str(
                        var,
                        eqn,
                        all_ic_bc_constants_str,
                        all_ic_bc_julia_str,
                        "boundary condition",
                    )

                    if side == "left":
                        limit = var_to_ind_vars_left_boundary[var.id]
                    elif side == "right":
                        limit = var_to_ind_vars_right_boundary[var.id]

                    bc = f"{variable_id_to_print_name[var.id]}{limit}"
                    if typ == "Dirichlet":
                        bc = bc
                    elif typ == "Neumann":
                        bc = f"D{domain_name_to_symbol[tuple(var.domain)]}({bc})"
                    all_ic_bc_str += f"   {bc} ~ {eqn_str},\n"

    # Replace variables in the julia strings that correspond to pybamm variables with
    # their julia equivalent
    for var_id, julia_id in variable_id_to_print_name.items():
        # e.g. boundary_value_right(cache_123456789) gets replaced with u1(t, 1)
        cache_var_id = id_to_julia_variable(var_id, "cache")
        if f"boundary_value_right({cache_var_id})" in all_ic_bc_julia_str:
            all_ic_bc_julia_str = all_ic_bc_julia_str.replace(
                f"boundary_value_right({cache_var_id})",
                julia_id + var_to_ind_vars_right_boundary[var_id],
            )
        # e.g. cache_123456789 gets replaced with u1(t, x)
        all_ic_bc_julia_str = all_ic_bc_julia_str.replace(
            cache_var_id, julia_id + var_to_ind_vars[var_id]
        )

    ####################################################################################

    # Create ODESystem or PDESystem
    if not is_pde:
        mtk_str += "sys = ODESystem(eqs, t)\n\n"
        mtk_str += (
            all_ic_bc_constants_str
            + all_ic_bc_julia_str
            + "\n"
            + f"u0 = [\n{all_ic_bc_str}]\n"
        )
    else:
        # Initial and boundary conditions
        mtk_str += (
            all_ic_bc_constants_str
            + all_ic_bc_julia_str
            + "\n"
            + f"ics_bcs = [\n{all_ic_bc_str}]\n"
        )

        # Domains
        mtk_str += "\n"
        tpsan_str = ",".join(
            map(lambda x: f"{x / model.timescale.evaluate():.3f}", tspan)
        )
        mtk_str += f"t_domain = Interval({tpsan_str})\n"
        domains = "domains = [\n   t in t_domain,\n"
        for domain, symbol in domain_name_to_symbol.items():
            limits = domain_name_to_limits[tuple(domain)]
            if limits is not None:
                mtk_str += f"{symbol}_domain = Interval{limits}\n"
                domains += f"   {symbol} in {symbol}_domain,\n"
        domains += "]\n"

        mtk_str += "\n"
        mtk_str += domains

        # Independent and dependent variables
        mtk_str += "ind_vars = [{}]\n".format(", ".join(ind_vars))
        mtk_str += "dep_vars = [{}]\n\n".format(", ".join(dep_vars))

        name = model.name.replace(" ", "_").replace("-", "_")
        mtk_str += (
            name
            + "_pde_system = PDESystem(eqs, ics_bcs, domains, ind_vars, dep_vars)\n\n"
        )

    # Replace parameters in the julia strings in the form "inputs[name]"
    # with just "name"
    for param in model.input_parameters:
        mtk_str = mtk_str.replace(f"inputs['{param.name}']", param.name)

    # Need to add 'nothing' to the end of the mtk string to avoid errors in MTK v4
    # See https://github.com/SciML/diffeqpy/issues/82
    mtk_str += "nothing\nend\n"

    return mtk_str



#
# Geometric Parameters
#
import pybamm
from .base_parameters import BaseParameters


class GeometricParameters(BaseParameters):
    """
    Standard geometric parameters

    Layout:
        1. Dimensional Parameters
        2. Dimensional Functions
        3. Scalings
        4. Dimensionless Parameters
        5. Dimensionless Functions
    """

    def __init__(self):
        self.n = DomainGeometricParameters("Negative", self)
        self.s = DomainGeometricParameters("Separator", self)
        self.p = DomainGeometricParameters("Positive", self)
        self.domain_params = [self.n, self.s, self.p]

        # Set parameters and scales
        self._set_dimensional_parameters()
        self._set_scales()
        self._set_dimensionless_parameters()

    def _set_dimensional_parameters(self):
        """Defines the dimensional parameters."""
        for domain in self.domain_params:
            domain._set_dimensional_parameters()

        # Macroscale geometry
        self.L_x = (
            self.n.L + self.s.L + self.p.L
        )  # Total distance between current collectors
        self.L = self.n.L_cc + self.L_x + self.p.L_cc  # Total cell thickness
        self.L_Li = pybamm.Parameter("Lithium counter electrode thickness [m]")
        self.L_y = pybamm.Parameter(
            "Electrode width [m]"
        )  # For a cylindrical cell L_y is the "unwound" length of the electrode
        self.L_z = pybamm.Parameter("Electrode height [m]")
        self.r_inner_dimensional = pybamm.Parameter("Inner cell radius [m]")
        self.r_outer_dimensional = pybamm.Parameter("Outer cell radius [m]")
        self.A_cc = self.L_y * self.L_z  # Current collector cross sectional area
        self.A_cooling = pybamm.Parameter("Cell cooling surface area [m2]")
        self.V_cell = pybamm.Parameter("Cell volume [m3]")

    def _set_scales(self):
        """Define the scales used in the non-dimensionalisation scheme"""
        for domain in self.domain_params:
            domain._set_scales()

    def _set_dimensionless_parameters(self):
        """Defines the dimensionless parameters."""
        for domain in self.domain_params:
            domain._set_dimensionless_parameters()

        # Macroscale Geometry
        self.l_x = self.L_x / self.L_x
        self.l_Li = self.L_Li / self.L_x
        self.l_y = self.L_y / self.L_z
        self.l_z = self.L_z / self.L_z
        self.r_inner = self.r_inner_dimensional / self.r_outer_dimensional
        self.r_outer = self.r_outer_dimensional / self.r_outer_dimensional
        self.a_cc = self.l_y * self.l_z
        self.a_cooling = self.A_cooling / (self.L_z ** 2)
        self.v_cell = self.V_cell / (self.L_x * self.L_z ** 2)

        self.l = self.L / self.L_x
        self.delta = self.L_x / self.L_z  # Pouch cell aspect ratio


class DomainGeometricParameters(BaseParameters):
    def __init__(self, domain, main_param):
        self.domain = domain
        self.main_param = main_param

    def _set_dimensional_parameters(self):
        """Defines the dimensional parameters."""
        Domain = self.domain

        if self.domain == "Separator":
            self.L = pybamm.Parameter("Separator thickness [m]")
            self.b_e = pybamm.Parameter("Separator Bruggeman coefficient (electrolyte)")
            return

        # Macroscale geometry
        self.L_cc = pybamm.Parameter(f"{Domain} current collector thickness [m]")
        self.L = pybamm.Parameter(f"{Domain} electrode thickness [m]")

        # Tab geometry (for pouch cells)
        self.L_tab = pybamm.Parameter(f"{Domain} tab width [m]")
        self.Centre_y_tab = pybamm.Parameter(f"{Domain} tab centre y-coordinate [m]")
        self.Centre_z_tab = pybamm.Parameter(f"{Domain} tab centre z-coordinate [m]")
        self.A_tab = self.L_tab * self.L_cc  # Area of tab

        # Microscale geometry
        # Note: for li-ion cells, the definition of the surface area to
        # volume ratio is overwritten in lithium_ion_parameters.py to be computed
        # based on the assumed particle shape
        self.a_dim = pybamm.Parameter(
            f"{Domain} electrode surface area to volume ratio [m-1]"
        )
        self.b_e = pybamm.Parameter(
            f"{Domain} electrode Bruggeman coefficient (electrolyte)"
        )
        self.b_s = pybamm.Parameter(
            f"{Domain} electrode Bruggeman coefficient (electrode)"
        )

        # Particle-size distribution geometry
        self.R_min_dim = pybamm.Parameter(f"{Domain} minimum particle radius [m]")
        self.R_max_dim = pybamm.Parameter(f"{Domain} maximum particle radius [m]")
        self.sd_a_dim = pybamm.Parameter(
            f"{Domain} area-weighted particle-size standard deviation [m]"
        )

    @property
    def R_dimensional(self):
        if self.domain == "Negative":
            x = pybamm.standard_spatial_vars.x_n
        elif self.domain == "Positive":
            x = pybamm.standard_spatial_vars.x_p

        return pybamm.FunctionParameter(
            f"{self.domain} particle radius [m]",
            {"Through-cell distance (x) [m]": x * self.main_param.L_x},
        )

    def f_a_dist_dimensional(self, R):
        """
        Dimensional electrode area-weighted particle-size distribution
        """
        inputs = {f"{self.domain} particle-size variable [m]": R}
        return pybamm.FunctionParameter(
            f"{self.domain} area-weighted particle-size distribution [m-1]", inputs
        )

    def _set_scales(self):
        """Define the scales used in the non-dimensionalisation scheme"""
        if self.domain == "Separator":
            return
        # Microscale geometry
        # Note: these scales are necessary here to non-dimensionalise the
        # particle size distributions.
        self.R_typ = pybamm.xyz_average(self.R_dimensional)

    def _set_dimensionless_parameters(self):
        """Defines the dimensionless parameters."""
        main = self.main_param

        # Macroscale Geometry
        self.l = self.L / main.L_x
        if self.domain == "Separator":
            return

        self.l_cc = self.L_cc / main.L_x

        # Tab geometry (for pouch cells)
        self.l_tab = self.L_tab / main.L_z
        self.centre_y_tab = self.Centre_y_tab / main.L_z
        self.centre_z_tab = self.Centre_z_tab / main.L_z

        # Particle-size distribution geometry
        self.R_min = self.R_min_dim / self.R_typ
        self.R_max = self.R_max_dim / self.R_typ
        self.sd_a = self.sd_a_dim / self.R_typ

        # Particle radius
        self.R = self.R_dimensional / self.R_typ

    def f_a_dist(self, R):
        """
        Dimensionless electrode area-weighted particle-size distribution
        """
        R_dim = R * self.R_typ
        return self.f_a_dist_dimensional(R_dim) * self.R_typ


geometric_parameters = GeometricParameters()



#
# Standard thermal parameters
#
import pybamm
from .base_parameters import BaseParameters


class ThermalParameters(BaseParameters):
    """
    Standard thermal parameters

    Layout:
        1. Dimensional Parameters
        2. Dimensional Functions
        3. Dimensionless Parameters
        4. Dimensionless Functions
    """

    def __init__(self):

        # Get geometric parameters
        self.geo = pybamm.geometric_parameters

        self.n = DomainThermalParameters("Negative", self)
        self.s = DomainThermalParameters("Separator", self)
        self.p = DomainThermalParameters("Positive", self)
        self.domain_params = [self.n, self.s, self.p]

        # Set parameters
        self._set_dimensional_parameters()
        self._set_dimensionless_parameters()

    def _set_dimensional_parameters(self):
        """Defines the dimensional parameters"""
        for domain in self.domain_params:
            domain._set_dimensional_parameters()

        # Reference temperature
        self.T_ref = pybamm.Parameter("Reference temperature [K]")

        # Cooling coefficient
        self.h_edge_dim = pybamm.Parameter("Edge heat transfer coefficient [W.m-2.K-1]")
        self.h_total_dim = pybamm.Parameter(
            "Total heat transfer coefficient [W.m-2.K-1]"
        )

        # Typical temperature rise
        self.Delta_T = pybamm.Scalar(1)

        # Initial temperature
        self.T_init_dim = pybamm.Parameter("Initial temperature [K]")

        # References
        self.rho_eff_dim_ref = self.rho_eff_dim(self.T_ref)
        self.lambda_eff_dim_ref = self.lambda_eff_dim(self.T_ref)

        # Planar (y,z) thermal diffusion timescale
        self.tau_th_yz = (
            self.rho_eff_dim_ref * (self.geo.L_z ** 2) / self.lambda_eff_dim_ref
        )

    def T_amb_dim(self, t):
        """Dimensional ambient temperature"""
        return pybamm.FunctionParameter("Ambient temperature [K]", {"Time [s]": t})

    def rho_eff_dim(self, T):
        """Effective volumetric heat capacity [J.m-3.K-1]"""
        return (
            self.n.rho_cc_dim(T) * self.n.c_p_cc_dim(T) * self.geo.n.L_cc
            + self.n.rho_dim(T) * self.n.c_p_dim(T) * self.geo.n.L
            + self.s.rho_dim(T) * self.s.c_p_dim(T) * self.geo.s.L
            + self.p.rho_dim(T) * self.p.c_p_dim(T) * self.geo.p.L
            + self.p.rho_cc_dim(T) * self.p.c_p_cc_dim(T) * self.geo.p.L_cc
        ) / self.geo.L

    def lambda_eff_dim(self, T):
        """Effective thermal conductivity [W.m-1.K-1]"""
        return (
            self.n.lambda_cc_dim(T) * self.geo.n.L_cc
            + self.n.lambda_dim(T) * self.geo.n.L
            + self.s.lambda_dim(T) * self.geo.s.L
            + self.p.lambda_dim(T) * self.geo.p.L
            + self.p.lambda_cc_dim(T) * self.geo.p.L_cc
        ) / self.geo.L

    def _set_dimensionless_parameters(self):
        """Defines the dimensionless parameters"""
        for domain in self.domain_params:
            domain._set_dimensionless_parameters()

        # Relative temperature rise
        self.Theta = self.Delta_T / self.T_ref

        # Cooling coefficient
        self.h_edge = self.h_edge_dim * self.geo.L_x / self.lambda_eff_dim_ref
        self.h_total = self.h_total_dim * self.geo.L_x / self.lambda_eff_dim_ref

        # Initial temperature
        self.T_init = (self.T_init_dim - self.T_ref) / self.Delta_T

    def rho(self, T):
        """
        Dimensionless effective density, not to be confused with rho_eff_dim,
        which is the dimensional effective volumetric heat capacity
        """
        return (
            self.n.rho_cc(T) * self.geo.n.l_cc
            + self.n.rho(T) * self.geo.n.l
            + self.s.rho(T) * self.geo.s.l
            + self.p.rho(T) * self.geo.p.l
            + self.p.rho_cc(T) * self.geo.p.l_cc
        ) / self.geo.l

    def T_amb(self, t):
        """Dimensionless ambient temperature"""
        return (self.T_amb_dim(t) - self.T_ref) / self.Delta_T


class DomainThermalParameters(BaseParameters):
    def __init__(self, domain, main_param):
        self.domain = domain
        self.main_param = main_param

    def _set_dimensional_parameters(self):
        self.h_cc_dim = pybamm.Parameter(
            f"{self.domain} current collector surface heat transfer coefficient "
            "[W.m-2.K-1]"
        )
        self.h_tab_dim = pybamm.Parameter(
            f"{self.domain} tab heat transfer coefficient [W.m-2.K-1]"
        )

    def c_p_dim(self, T):
        """Electrode specific heat capacity [J.kg-1.K-1]"""
        inputs = {"Temperature [K]": T}
        if self.domain == "Separator":
            name = "Separator specific heat capacity [J.kg-1.K-1]"
        else:
            name = f"{self.domain} electrode specific heat capacity [J.kg-1.K-1]"
        return pybamm.FunctionParameter(name, inputs)

    def c_p_cc_dim(self, T):
        """Current collector specific heat capacity [J.kg-1.K-1]"""
        inputs = {"Temperature [K]": T}
        return pybamm.FunctionParameter(
            f"{self.domain} current collector specific heat capacity [J.kg-1.K-1]",
            inputs,
        )

    def lambda_dim(self, T):
        """Electrode thermal conductivity [W.m-1.K-1]"""
        inputs = {"Temperature [K]": T}
        if self.domain == "Separator":
            name = "Separator thermal conductivity [W.m-1.K-1]"
        else:
            name = f"{self.domain} electrode thermal conductivity [W.m-1.K-1]"
        return pybamm.FunctionParameter(name, inputs)

    def lambda_cc_dim(self, T):
        """Current collector thermal conductivity [W.m-1.K-1]"""
        inputs = {"Temperature [K]": T}
        return pybamm.FunctionParameter(
            f"{self.domain} current collector thermal conductivity [W.m-1.K-1]", inputs
        )

    def rho_dim(self, T):
        """Electrode density [kg.m-3]"""
        inputs = {"Temperature [K]": T}
        if self.domain == "Separator":
            name = "Separator density [kg.m-3]"
        else:
            name = f"{self.domain} electrode density [kg.m-3]"
        return pybamm.FunctionParameter(name, inputs)

    def rho_cc_dim(self, T):
        """Current collector density [kg.m-3]"""
        inputs = {"Temperature [K]": T}
        return pybamm.FunctionParameter(
            f"{self.domain} current collector density [kg.m-3]", inputs
        )

    def _set_dimensionless_parameters(self):
        main = self.main_param
        self.h_cc = self.h_cc_dim * main.geo.L_x / main.lambda_eff_dim_ref
        self.h_tab = self.h_tab_dim * main.geo.L_x / main.lambda_eff_dim_ref

    def rho(self, T):
        """Dimensionless electrode density"""
        T_dim = self.main_param.Delta_T * T + self.main_param.T_ref
        return (
            self.rho_dim(T_dim) * self.c_p_dim(T_dim) / self.main_param.rho_eff_dim_ref
        )

    def rho_cc(self, T):
        """Dimensionless current collector density"""
        T_dim = self.main_param.Delta_T * T + self.main_param.T_ref
        return (
            self.rho_cc_dim(T_dim)
            * self.c_p_cc_dim(T_dim)
            / self.main_param.rho_eff_dim_ref
        )

    def lambda_(self, T):  # cannot call a function "lambda"
        """Dimensionless electrode thermal conductivity"""
        T_dim = self.main_param.Delta_T * T + self.main_param.T_ref
        return self.lambda_dim(T_dim) / self.main_param.lambda_eff_dim_ref

    def lambda_cc(self, T):
        """Dimensionless current collector thermal conductivity"""
        T_dim = self.main_param.Delta_T * T + self.main_param.T_ref
        return self.lambda_cc_dim(T_dim) / self.main_param.lambda_eff_dim_ref


thermal_parameters = ThermalParameters()



#
# Base parameters class
#
import pybamm


class BaseParameters:
    """
    Overload the `__setattr__` method to record what the variable was called.
    """

    def __getattribute__(self, name):
        """
        Raise more informative error to users when they try to access a
        non-existent attribute, which may have recently changed name
        """
        try:
            return super().__getattribute__(name)
        except AttributeError as e:
            for domain in ["n", "s", "p"]:
                if f"_{domain}_" in name or name.endswith(f"_{domain}"):
                    name_without_domain = name.replace(f"_{domain}_", "_").replace(
                        f"_{domain}", ""
                    )
                    raise AttributeError(
                        f"param.{name} does not exist. It may have been renamed to "
                        f"param.{domain}.{name_without_domain}"
                    )
            raise e

    def __setattr__(self, name, value):
        if hasattr(self, "domain"):
            d = self.domain.lower()[0]
            print_name = f"{name}_{d}"
        else:
            print_name = name
        if isinstance(value, pybamm.Symbol):
            value.print_name = print_name
        super().__setattr__(name, value)
